' //////////////////////////////////////////////////////////////////////
' TITLE: RJH_SpaceWar_Virtual_VGA_032.SPIN - Free flight space battle
'
' DESCRIPTION:
'          Based on the orinial PDP-1 SpaceWar! by Steve "Slug" Russell, Martin "Shag" Graetz and Wayne Wiitanen,
'          with modifications based on other implementations, and whatever struck my fancy.
'          SpaceWar! was hacked relentlessly by the development community in its day.  Feel free to 
'          continue the tradition and hack away!  Send me your cool mods and I'll work them into
'          the main release.
'
' AUTHOR:  Eric Moyer
'          Monumental thanks to Andre' LaMothe, for creating "SPACE_DEMO_001", on which
'          this game is based.
'
'          VGA Modifications by Ross Higson.
'
' UPDATES: Search the Parallax forums at http://forums.parallax.com for the latest version.
'
' CONTROLS:
'          Two NES gamepads, or a gamepad and the keyboard
'          Press START or <Spacebar> at the GAME OVER screen to return to the options menu.
'
'          Player 0    Player 1       Function 
'          --------    ------------   ------------------------------------------------------------------------
'          LEFT        LEFT  or 'A'   Rotate Left
'          RIGHT       RIGHT or 'D'   Rotate Right
'          UP          UP    or 'W'   Shields       
'          DOWN        DOWN  or 'S'   Hyperspace
'          A           A     or 'Y'   Fire
'          B           B     or 'G'   Thrust
'
'          The options screen is controlled by player 0's gamepad.  Use UP/DOWN to choose, and A to toggle an option.
'
' HACKING: If you want to hack SpaceWar!, start by turning on DEBUG__SHOW_PACING_OVERRUNS. If your processing
'          overhead overruns the allotted frame rate a white square will appear in the upper right hand corner
'          for each overrun frame.
'
'          If you want to monitor a variable, turn on DEBUG__SHOW_WATCH_VARIABLES and set watchval_fixed or 
'          watchval_int (for fixed point and integer values) in the main execution loop somewhere.  Monitoring
'          will slow down the frame rate a bit.
'
'          If you set DEBUG__ENABLE_TEST_MONKEY the game will play itself randomly. Usefull
'          for getting a long-term "burn in" run to verify code stability.
'                                                                   
'
' //////////////////////////////////////////////////////////////////////
'
'  Revision History
'
'  Rev  Date      Description
'  ---  --------  ---------------------------------------------------
'  011  04-13-07  First playable release.
'  013  04-25-07  Release for first upload.
'  014  05-01-07  Migrate to 2 color video driver to open up 12K of code space.
'                 Add keyboard control for player 1.
'                 Fix pacing so that frame rate is consistent regardless of the processing overhead
'                   within a given frame.
'                 Add shield animation.
'                 Add explode, disappear, delay sequencing to player death.
'                 Use temporary debug graphic ('x') for exploding ship.
'                 Add ship/ship collision detection.
'                 Add shield functionality (against shot/ship and ship/ship).
'                 Show shield energy, and deplete as used.
'                 Shrink shots to single pixels.
'  015  05-05-07  Add Nick Sabalausky's sound driver (not using any sounds yet though).
'                 Add fuel tracking and display fuel gauge.
'                 Cleanup/optimize player stats display code.
'                 Implement explosion particle animation.
'                 Modify pacing so that waitcnt() call is bypassed whenever game frames run
'                    over their allotted time.
'                 Add options screen, and change gameflow so next match can be started
'                    without resetting the console.
'                 Implement option flags for all currently supported options.
'                 Move origination point of shots to nose of ship (was center).
'                 Use fixed point math for shield and fuel tracking to avoid a costly divide when
'                    rendering the gauges.
'                 Swapped shield/fuel so shields are the dashed line.
'                 Offset thrust drawing to rear of ships.
'                *Temporarily disable shot/shield bounce code (not ready yet).
'                 Fix pacing calculation to correct pacing hang
'                 Fix bug where shield or fuel could run negative
'  016  05-06-07  Code some of BFG for testing (not ready for release yet).
'                 Add debug mode.
'                 Change option menu up/down behavior to work on keydown event, and to autorepeat.
'                 Change default options, and make non-inertial shots the default setting.
'                 Increase the velocity of non-inertial shots.
'                 Change black hole animation
'  017  05-07-07  Revise pacing algorithm - found a flaw in the old one.
'                 Light debug LED during pacing delay.
'                 Add test monkey.
'                *DEBUG__SHOW_PACING_OVERRUNS is temporarily non-functional.  Will fix when I port
'                   the new pacing algorithm to assembly where it belongs.
'  018  05-15-07  Fix bug where keyboard control did not work unless controller 1 was plugged in
'                   (which embarassingly defeated the whole point of adding keyboard control!)
'  019  05-15-07  Add upper EEPROM memory asset managment.
'                 Add Nick's EEPROM driver.
'                 Load starup screen from EEPROM
'  022  06-20-07  Disable EEPROM stuff for now.
'                 Fix game over hang (ignoring start button) when only one connector is inserted.
'                 Add sound effects.
'  023  06-20-07  Fix indexing bug on player shots.
'
'  VGA Version split off at version 025:
'
'  030  07-27-08  VGA Version - up to 1152x864 (2 colour) or 800x600 (4 colour)
'                 Added ships reminiscent of Cinematronics version
'                 Added stars (and twinkle option)
'                 Added shield bounce
'                 Added non-determinism in shot time to live and shield bounce
'                 Added hyperspace option
'                 Added ship damage option
'  031  07-28-08  Added bounce back option
'                 Added negative gravity option
'                 Added ram recyling
'                 Added random rocks option 
'  032  08-01-08  Shots were not being absorbed properly by rocks
'                 Added expanded universe option - universe is 25% bigger than screen on each side
'                 Renamed 'closed universe' option as 'bounce back' (as per cinematronics version)
'                 Fixed timing problem in Virtual VGA driver on startup at high resolution
'
'                 
'  Open Issues:  
'     High Priority:
'
'     Medium Priority:
'       Modify explosion particle distribution to a circular pattern with highest density nearest the origin.
' 
'     Lower Priority:
'       Add remaining options: 
'          BFG.
'       Properly handle dual death (wait for both players to sequence through explode, disappear, delay) 
'          before starting the next match
'
'       Possible other options:
'         Handicap (vary starting number of ships for each player)
'
'
' //////////////////////////////////////////////////////////////////////
'
' SCCuM Preprocessor Definitions
' (You can Get SCCuM here: http://forums.parallax.com/forums/default.aspx?f=33&m=194963)
'
'#DEFINE LCD_DEBUGGER_ENABLE    = FALSE
'
' Don't bother turning either of these options on, the EEPROM graphics and music assets have not been released
'#DEFINE EEPROM_GRAPHICS_ENABLE = TRUE   
'#DEFINE MUSIC_ENABLE           = TRUE
'
'
'///////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

CON

'--- USER CONFIGURATION -----------------------------------------------------------------------------------------------
' Uncomment one of the sections below to select your resolution. Also select your desired COLOUR_MODE
'
' Supported resolutions are:
'
'  640x480 2 colour & 4 colour
'  800x600 2 colour & 4 colour
' 1024x768 2 colour only
' 1152x864 2 colour only
'

  COLOUR_MODE       = 0                ' 0 = 2 colour mode, 1 = 4 colour mode. Not all resolutions support 4 colours.

  DOUBLE_BUFFER     = TRUE             ' Double buffering is recommended for screens that are redrawn rapidly.
                                       ' (Note that you cannot turn double buffering on or off "on the fly").
                                       ' It is recommended to leave this on.

  RAM_RECYCLE       = TRUE             ' Reclaim the RAM used for cog data once the cogs are loaded. If you set
                                       ' this to FALSE you will probably not be able to have sound effects or
                                       ' keyboard support at the higher resolutions (*)                                       

  SFX_COG           = TRUE             ' SFX Manager runs on own cog (better quality sound). Set this to false if
                                       ' you want sound effects AND the keyboard - otherwise you run out of cogs.

' (* Note: to simplify disabling sound effects or keyboard support, dummy drivers are provided. See the OBJ section)

'--- 640 x 480 --------------------------------------------------------------------------------------------------------
{
  ' physical modeling constants:
  SPACE_FRICTION                = 8    ' from 0 to 15, 0 is infinite friction, 15 is 0.000001 approximately
  MAX_SHIP_VEL                  = 25   ' maximum velocity of player's ship
  SAR_SHIP_VEL                  = 3    ' sar by this amount
  MAX_PROJECTILE_VEL            = 80
  SCL_PROJECTILE_VEL            = 6    ' scale by this amount
  ANG_ROTATE                    = 1
  SHL_ROTATE                    = 8    ' shl by this amount
  SHOT_TTL_INIT                 = 60   ' Shot time to live
  BLACK_HOLE_GRAVITY            = $01F0_0000
  GRAVITY_CLIP                  = $0004_0000
  SHIP_SCALE                    = $0020 '640 x 480 or less
  P0_THRUSTER_DISTANCE          = 5
  P1_THRUSTER_DISTANCE          = 11
  ROR_THRUST                    = 10    ' thrust ror by this amount
  COLLISION_RANGE_FP            = $0008_0000    ' Collision bounding box range, in fixed point
  SHIP_GUN_PLACEMENT_DISTANCE   = 8  
  NUM_STARS                     = 10
  NUM_DAMAGE_PARTICLES          = 3     ' particles for each item of damage
  DAMAGE_TTL_INIT               = 10    ' time to live for damage particles
  

  'vga constants:
  mode = COLOUR_MODE << 3 + %000
  '    72Hz       60Hz
  hd = 640        '640
  hf = 24         '16
  hs = 40         '96
  hb = 128        '48
  vd = 480        '480
  vf = 9          '10
  vs = 3          '4
  vb = 28         '33
  px = 31_500_000 '25_000_000
}
'--- 800 x 600 --------------------------------------------------------------------------------------------------------
{

  'physical modeling constants:
  SPACE_FRICTION                = 12   ' from 0 to 15, 0 is infinite friction, 15 is 0.000001 approximately
  MAX_SHIP_VEL                  = 30   ' maximum velocity of player's ship
  SAR_SHIP_VEL                  = 3    ' sar by this amount 
  MAX_PROJECTILE_VEL            = 100
  SCL_PROJECTILE_VEL            = 10    ' scale by this amount
  ANG_ROTATE                    = 1
  SHL_ROTATE                    = 8    ' shl by this amount
  SHOT_TTL_INIT                 = 45   ' Shot time to live
  BLACK_HOLE_GRAVITY            = $0320_0000
  GRAVITY_CLIP                  = $0006_0000
  SHIP_SCALE                    = $0030
  P0_THRUSTER_DISTANCE          = 7
  P1_THRUSTER_DISTANCE          = 16
  ROR_THRUST                    = 10    ' thrust ror by this amount
  COLLISION_RANGE_FP            = $000E_0000    ' Collision bounding box range, in fixed point
  SHIP_GUN_PLACEMENT_DISTANCE   = 13            ' 
  NUM_STARS                     = 15
  NUM_DAMAGE_PARTICLES          = 3     ' particles for each item of damage
  DAMAGE_TTL_INIT               = 15    ' time to live for damage particles
  

  'vga constants:
  mode = COLOUR_MODE << 3 + %011
  hd = 800
  hf = 40
  hs = 128
  hb = 88
  vd = 600
  vf = 1
  vs = 4
  vb = 23
  px = $2625A00 '40_000_000
}

'--- 1024 x 768 -------------------------------------------------------------------------------------------------------
'{
  ' physical modeling constants:
  SPACE_FRICTION                = 13   ' from 0 to 15, 0 is infinite friction, 15 is 0.000001 approximately
  MAX_SHIP_VEL                  = 40   ' maximum velocity of player's ship
  SAR_SHIP_VEL                  = 3    ' sar by this amount 
  MAX_PROJECTILE_VEL            = 100
  SCL_PROJECTILE_VEL            = 12    ' scale by this amount
  ANG_ROTATE                    = 1
  SHL_ROTATE                    = 8    ' shl by this amount
  SHOT_TTL_INIT                 = 50   ' Shot time to live
  BLACK_HOLE_GRAVITY            = $0560_0000
  GRAVITY_CLIP                  = $0006_0000
  SHIP_SCALE                    = $0030
  P0_THRUSTER_DISTANCE          = 7
  P1_THRUSTER_DISTANCE          = 16
  ROR_THRUST                    = 10    ' thrust ror by this amount
  COLLISION_RANGE_FP            = $000E_0000    ' Collision bounding box range, in fixed point
  SHIP_GUN_PLACEMENT_DISTANCE   = 13  
  NUM_STARS                     = 15
  NUM_DAMAGE_PARTICLES          = 3     ' particles for each item of damage
  DAMAGE_TTL_INIT               = 20    ' time to live for damage particles

  'vga constants:
  mode = COLOUR_MODE << 3 + %000
  hd = 1024
  hf = 24
  hs = 136
  hb = 160
  vd = 768
  vf = 3
  vs = 6
  vb = 29
  px = 64_000_000      'reduces jitter - correct value is 65_000_000 

'}
'--- 1152 x 864 -------------------------------------------------------------------------------------------------------
{
  ' physical modeling constants:
  SPACE_FRICTION                = 13   ' from 0 to 15, 0 is infinite friction, 15 is 0.000001 approximately
  MAX_SHIP_VEL                  = 30   ' maximum velocity of player's ship
  SAR_SHIP_VEL                  = 3    ' sar by this amount 
  MAX_PROJECTILE_VEL            = 100
  SCL_PROJECTILE_VEL            = 8    ' scale by this amount
  ANG_ROTATE                    = 1
  SHL_ROTATE                    = 8    ' shl by this amount
  SHOT_TTL_INIT                 = 80   ' Shot time to live
  BLACK_HOLE_GRAVITY            = $0700_0000
  GRAVITY_CLIP                  = $0006_0000
  SHIP_SCALE                    = $0030
  P0_THRUSTER_DISTANCE          = 7
  P1_THRUSTER_DISTANCE          = 16
  ROR_THRUST                    = 10    ' thrust ror by this amount
  COLLISION_RANGE_FP            = $000E_0000    ' Collision bounding box range, in fixed point
  SHIP_GUN_PLACEMENT_DISTANCE   = 13 
  NUM_STARS                     = 15
  NUM_DAMAGE_PARTICLES          = 3     ' particles for each item of damage
  DAMAGE_TTL_INIT               = 20    ' time to live for damage particles

  'vga constants:
  mode = COLOUR_MODE << 3 + %001
  hd = 1152
  hf = 64
  hs = 112
  hb = 256
  vd = 864
  vf = 6
  vs = 5
  vb = 41
  px = 80_000_000
}
'----------------------------------------------------------------------------------------------------------------------


  _clkmode = xtal2 + pll8x            ' enable external clock and pll times 8
  _xinfreq = 10_000_000 + 0000        ' set frequency to 10 MHZ plus some error for xtals that are not exact add 1000-5000 usually works
  _stack = $32                       ' accomodate stack

  'DEBUG OPTIONS:
  DEBUG__SHOW_WATCH_VARIABLES   = FALSE
  DEBUG__SHOW_PACING_OVERRUNS   = FALSE
  DEBUG__ENABLE_DEBUG_MODE      = FALSE
  DEBUG__ENABLE_TEST_MONKEY     = FALSE
  DEBUG__SHOW_PACING_LED        = FALSE

  'EEPROM ASSET MAP
  EEPROM_SCREEN_TITLE           = $08000
  EEPROM_SCREEN_OPTIONS         = $09800 
  EEPROM_SONG_GOTHIC            = $0B000
  EEPROM_SONG_FF1_BATTLE        = $0BE52
  EEPROM_SONG_NORTH_WALL        = $10108
  EEPROM_SONG_GENTLE_WIND       = $117FB
  EEPROM_SONG_LOST_LOVE         = $122B8
  
  ' graphics driver and screen constants
  PARAMCOUNT        = 21        

  ' Status display constants 
  PLAYER_1_STATS_X_POS       = -SCREEN_WIDTH/2  + 10
  PLAYER_1_STATS_Y_POS       =  SCREEN_HEIGHT/2 - 12

  PLAYER_0_STATS_X_POS       =  SCREEN_WIDTH/2  - 10/2*12 - 8
  PLAYER_0_STATS_Y_POS       =  SCREEN_HEIGHT/2 - 12

  STATUS_GAUGE_LENGTH        = 50

  ' game states
  GAME_STATE_INIT      = 0
  GAME_STATE_MENU      = 1
  GAME_STATE_START     = 2
  GAME_STATE_RUN       = 3

  ' angular constants to make object declarations easier
  ANG_0    = $0000
  ANG_360  = $2000
  ANG_240  = ($2000*2/3)
  ANG_180  = ($2000/2)
  ANG_270  = ($2000*3/4)
  ANG_300  = ($2000*5/6)
  ANG_120  = ($2000/3)
  ANG_90   = ($2000/4)
  ANG_60   = ($2000/6)
  ANG_45   = ($2000/8)
  ANG_30   = ($2000/12)
  ANG_22_5 = ($2000/16)
  ANG_15   = ($2000/24)
  ANG_10   = ($2000/36)
  ANG_5    = ($2000/72)
  ANG_1    = ($2000/360)

  ' constants for math functions
  SIN      = 0
  COS      = 1

  ' player states
  SHIP_STATE_ALIVE     = $80
  SHIP_STATE_EXPLODING = $7F
  SHIP_STATE_STRICKEN  = $70
  SHIP_STATE_DIE_DELAY = $20
  SHIP_STATE_DEAD      = $00

  ' rock states
  ROCK_STATE_PRESENT   = $80
  ROCK_STATE_EXPLODING = $7F
  ROCK_STATE_DIE_DELAY = $20
  ROCK_STATE_ABSENT    = $00


  ' control interface
  THRUST_BUTTON_ID = 1
  FIRE_BUTTON_ID   = 0


  ' screen constants:
  X_TILES           = hd / (32>>COLOUR_MODE) 
  Y_TILES           = vd / 16 

  SCREEN_WIDTH      = X_TILES * (32>>COLOUR_MODE)   
  SCREEN_HEIGHT     = Y_TILES * 16 

  ' NES bit encodings
  NES_RIGHT  = %00000001
  NES_LEFT   = %00000010
  NES_DOWN   = %00000100
  NES_UP     = %00001000
  NES_START  = %00010000
  NES_SELECT = %00100000
  NES_B      = %01000000
  NES_A      = %10000000

  ' Number of players
  NUM_PLAYERS  = 2
  PLAYER_0     = 0
  PLAYER_1     = 1

  ' Player shots
  NUM_SHOTS = 12                              ' Number of shots on screen at any time    
  PLAYER_1_SHOT_START = (NUM_SHOTS / 2)       ' Index of first player 2 shot

  ' Random asteroids
  NUM_ROCKS = 2

  ' Objects
  NUM_OBJECTS            = NUM_PLAYERS + NUM_SHOTS + NUM_ROCKS
  OBJECT_PLAYER_0        = 0
  OBJECT_PLAYER_1        = 1
  OBJECT_PLAYER_0_SHOT_0 = NUM_PLAYERS
  OBJECT_PLAYER_1_SHOT_0 = NUM_PLAYERS + PLAYER_1_SHOT_START
  OBJECT_ROCK_0          = NUM_PLAYERS + NUM_SHOTS 

  ' Black holes
  BLACK_HOLE_X          = 0
  BLACK_HOLE_Y          = 0
  
  ' Game options
  OPTION_BLACK_HOLE      = %00000000_00000000_00000000_00000001
  OPTION_SHOT_GRAVITY    = %00000000_00000000_00000000_00000010
  OPTION_INERTIAL_SHOTS  = %00000000_00000000_00000000_00000100
  OPTION_SHIELDS         = %00000000_00000000_00000000_00001000
  OPTION_FUEL            = %00000000_00000000_00000000_00010000
  OPTION_TWINKLE         = %00000000_00000000_00000000_00100000
  OPTION_HYPERSPACE      = %00000000_00000000_00000000_01000000
  OPTION_DAMAGE          = %00000000_00000000_00000000_10000000
  OPTION_ROCKS           = %00000000_00000000_00000001_00000000
  OPTION_BOUNCE          = %00000000_00000000_00000010_00000000
  OPTION_NEGATIVE        = %00000000_00000000_00000100_00000000
  OPTION_EXPANDED        = %00000000_00000000_00001000_00000000

  OPTION_MONKEY_IS_ALIVE = %10000000_00000000_00000000_00000000   ' Enables input banger test
  ' Future options: Barriers?

  'Ship Flags
  SHIP_FLAG_SHIELDS_UP  = %00000001
  SHIP_FLAG_THRUSTING   = %00000010
  SHIP_FLAG_HYPERSPACE  = %00000100

  'Menu Flags
  MENU_FLAG__A_HELD                 = %00000001
  MENU_FLAG__UP_DOWN_HELD           = %00000010
  MENU_FLAG__REPEAT_GO_TIME_MET     = %00000100
  MENU_FLAG__DO_REPEAT_CLICK        = %00001000

  'BFG Flags
  BFG_FLAG__OUT_OF_PLAY       = %00000001
  BFG_FLAG__PICKUP            = %00000010
  BFG_FLAG__HELD_BY_PLAYER_0  = %00000100
  BFG_FLAG__HELD_BY_PLAYER_1  = %00001000
  BFG_FLAG__DEPLOYED          = %00010000

  'BFG
  BFG_SPOKE_SIZE              = 50
  BFG_HUB_SIZE                = 11

  'Shield settings
  SHIELD_CONSUMPTION_RATE_FXP = (STATUS_GAUGE_LENGTH << 16) / 300    ' Shield consumption rate, fixed point
  SHIELD_RADIUS               = 10

  'Fuel settings
  FUEL_CONSUMPTION_RATE_FXP   = (STATUS_GAUGE_LENGTH << 16) / 600    ' Fuel consumption rate, fixed point
  SHIP_HYPERSPACE_TIME        = 25 

  'Explosions
  NUM_EXPLOSION_PARTICLES = 12      ' Number of particles in an explosion
  EXPLOSION_RADIUS        = 30      ' Explosion radius, in pixels

  'Default game options
  DEFAULT_OPTIONS       = OPTION_BLACK_HOLE | OPTION_SHOT_GRAVITY | OPTION_SHIELDS | OPTION_FUEL | OPTION_INERTIAL_SHOTS | OPTION_TWINKLE | OPTION_HYPERSPACE  | OPTION_DAMAGE | OPTION_ROCKS

  'Vector drawing commands
  VECTOR_MOVE           = $4000
  VECTOR_DRAW           = $8000
  VECTOR_END            = $0000

  'Game pacing tick
  PACING_TICK           = 1666_666 '1666_666   ' Target game frame rate
  PACING_MARGIN         =    3_000

  'Black hole
  NUM_BLACK_HOLE_PARTICLES = 10 

'///////////////////////////////////////////////////////////////////////
' VARIABLES SECTION ////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

VAR

  long  vga_status    '0/1/2 = off/visible/invisible           read-only
  long  vga_enable    '0/? = off/on                            write-only
  long  vga_pins      '%ppmmm = pins                           write-only
  long  vga_mode      '%ccinp = chroma,interlace,ntsc/pal,swap write-only
  long  vga_screen    'pointer to screen (words)               write-only
  long  vga_colors    'pointer to colors (longs)               write-only               
  long  vga_hc        'horizontal cells                        write-only
  long  vga_vc        'vertical cells                          write-only
  long  vga_hx        'horizontal cell expansion               write-only
  long  vga_vx        'vertical cell expansion                 write-only
  long  vga_ho        'horizontal offset                       write-only
  long  vga_vo        'vertical offset                         write-only
  long  vga_hd        'horizontal display pixels               write-only
  long  vga_hf        'horizontal front-porch pixels           write-only
  long  vga_hs        'horizontal sync pixels                  write-only
  long  vga_hb        'horizontal back-porch pixels            write-only
  long  vga_vd        'vertical display lines                  write-only
  long  vga_vf        'vertical front-porch lines              write-only
  long  vga_vs        'vertical sync lines                     write-only
  long  vga_vb        'vertical back-porch lines               write-only
  long  vga_rate      'pixel rate (Hz)                         write-only

  long  line_req
  long  line_buf


' RJH - implement 4 bits of colour -------------------------------------------------------------- OLD
'
  long  line_1[x_tiles]       ' (1 long per x tile - pixels only)
  long  line_2[x_tiles]       ' (1 long per x tile - pixels only)
  long  line_3[x_tiles]       ' (1 long per x tile - pixels only)
  long  line_4[x_tiles]       ' (1 long per x tile - pixels only)

' RJH - implement 4 bits of colour -------------------------------------------------------------- NEW
'
' NOTE : THIS DOESN'T WORK YET - NEED 2 VGA DRIVERS AS WELL AS 2 LINE GENERATORS !!!
'
'  long  line_1[x_tiles<<COLOUR_MODE]       ' (1 long per x tile - pixels only, or colours then pixels)
'  long  line_2[x_tiles<<COLOUR_MODE]       ' (1 long per x tile - pixels only, or colours then pixels)
'  long  line_3[x_tiles<<COLOUR_MODE]       ' (1 long per x tile - pixels only, or colours then pixels)
'  long  line_4[x_tiles<<COLOUR_MODE]       ' (1 long per x tile - pixels only, or colours then pixels)

' RJH - implement 4 bits of colour -------------------------------------------------------------- END

  long  initial_ram
  long  initial_tiles

  long  fast_screen[7]

  word  onscreen[x_tiles * y_tiles]  ' storage for screen tile map 
  word  offscreen[x_tiles * y_tiles] ' storage for screen tile map
  word  map[x_tiles * y_tiles]       ' storage for screen tile map

  long  colors[16]                   ' color look up table (16 combinations max for VGA - only 1 currently used)

  ' common object properties
  long object_x[NUM_OBJECTS]
  long object_y[NUM_OBJECTS]
  long object_dx[NUM_OBJECTS]
  long object_dy[NUM_OBJECTS]
  
  ' player's ship
  byte ship_state[NUM_PLAYERS]
  byte ship_flags[NUM_PLAYERS]
  long ship_shield_power_fxp[NUM_PLAYERS]
  long ship_fuel_fxp[NUM_PLAYERS] 
  byte lives[NUM_PLAYERS]
  byte damage[NUM_PLAYERS]
  long ship_angle[NUM_PLAYERS]
  long thrust_dx[NUM_PLAYERS], thrust_dy[NUM_PLAYERS] 
  long friction_dx[NUM_PLAYERS], friction_dy[NUM_PLAYERS]
  long bforce_dx[NUM_PLAYERS], bforce_dy[NUM_PLAYERS]
  long player_nes_buttons[NUM_PLAYERS]
  long score[NUM_PLAYERS]
  long shield_angle

  long player0_ship[4]
  long player1_ship[4]

  ' bfg
  long bfg_x
  long bfg_y
  byte bfg_flags

  ' rock
  LONG rock_state[NUM_ROCKS]                            ' present, exploding or absent 
  LONG rock_scale[NUM_ROCKS]                            ' floating point
  LONG rock_angle[NUM_ROCKS]
  LONG rock_spin[NUM_ROCKS]

  ' explosion particles
  long explosion_dx_fxp[NUM_EXPLOSION_PARTICLES]        ' particle delta x (fixed point)
  long explosion_dy_fxp[NUM_EXPLOSION_PARTICLES]        ' particle delta y (fixed point)

  ' damage particles
  long damage_x[NUM_DAMAGE_PARTICLES*4]                 ' particle x 
  long damage_y[NUM_DAMAGE_PARTICLES*4]                 ' particle y 
  long damage_dx[NUM_DAMAGE_PARTICLES*4]                ' particle delta x
  long damage_dy[NUM_DAMAGE_PARTICLES*4]                ' particle delta y
  byte damage_ttl[NUM_DAMAGE_PARTICLES*4]               ' particle time to live

  ' player's shots
  word shot_time_to_live[NUM_SHOTS]

  ' ship time to remain in hyperspace
  byte hyper_time[NUM_PLAYERS]

  ' stars
  long star_x[NUM_STARS]
  long star_y[NUM_STARS]

  long random_var                                         ' global random variable
  long curr_count                                         ' saves counter  

  byte fire_debounce                                      ' button debounce

  ' nes gamepad vars
  long nes_buttons
  long button_temp

  ' black hole
  long black_hole_particle_mask
  long black_hole_particle_angle[NUM_BLACK_HOLE_PARTICLES]
  long black_hole_particle_radius[NUM_BLACK_HOLE_PARTICLES] 

  ' game mode options
  long game_options

  ' pacing
  long pacing_cnt
  long pacing_target
  long pacing_wrapped_target
  
  ' sound effects process stack
  long sfx_stack[40]

  '#IF LCD_DEBUGGER_ENABLE
  { '<added by SCCuM preprocessor utility>
  byte debugger_key_held
  } '<added by SCCuM preprocessor utility>
  '#ENDIF
  
'///////////////////////////////////////////////////////////////////////
' OBJECT DECLARATION SECTION ///////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////
OBJ

  vga    : "RJH_Virtual_VGA_drv_012.spin"               ' Virtual VGA driver
  gr     : "RJH_Virtual_Graphics_drv_011.spin"          ' Virtual Graphics driver
   
  'key   : "keyboard_iso_011.spin"                      ' Parallax keyboard driver
  key   : "RJH_dummy_keyboard_011.spin"                    ' dummy driver - use this to disable keyboard
  
  'hdmf  : "RJH_HDMF_driver_012.spin"                   ' HDMF song player (for songs and sfx) N.B. CURRENTLY NOT ENOUGH COGS FOR EEPROM ASSETS! 
  hdmf  : "RJH_sound_drv_053_22kHz_16bit.spin"          ' sound driver (for sfx only)
  'hdmf  : "RJH_dummy_sound_011.spin"                   ' dummy driver - use this to disable sound effects                       
  

  '#IF LCD_DEBUGGER_ENABLE  
  { '<added by SCCuM preprocessor utility>
  debug  : "EPM_LCD_debug_monitor_001.spin"     ' LCD Debugger
  } '<added by SCCuM preprocessor utility>
  '#ENDIF  

'///////////////////////////////////////////////////////////////////////
' EXPORT PUBLICS  //////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

PUB start | i , j, dx, dy

  '//// GLOBAL INITIALIZTION
  
  'init debug LED
  DIRA[0] := 1 ' output 
  OUTA[0] := 0
  
  'default game options
  game_options := DEFAULT_OPTIONS
                                                                         
  if DEBUG__ENABLE_TEST_MONKEY
    game_options |= OPTION_MONKEY_IS_ALIVE
  
  'setup vga
  longmove(@vga_status, @vgaparams, paramcount)
  vga_colors := @colors

  line_req := 0
  line_buf := 0

  fast_screen[0] := @line_req
  fast_screen[1] := @line_buf
  if DOUBLE_BUFFER
    fast_screen[2] := @onscreen
  else
    fast_screen[2] := @offscreen
  fast_screen[3] := @line_1
  fast_screen[4] := @line_2
  fast_screen[5] := @line_3
  fast_screen[6] := @line_4
  vga_screen := @fast_screen

  player0_ship[0] := @player0_ship0
  player0_ship[1] := @player0_ship1
  player0_ship[2] := @player0_ship2
  player0_ship[3] := @player0_ship3
  player1_ship[0] := @player1_ship0
  player1_ship[1] := @player1_ship1
  player1_ship[2] := @player1_ship2
  player1_ship[3] := @player1_ship3

  'init colors - colour 1 is white so it can be used in both 2 colour and 4 colour mode
  repeat i from 0 to 16
    colors[i] := %11000000_00001100_11111100_00000000

  initial_ram := (word[$e] + _stack*4 + $00FF) & $FF00
  initial_tiles := ($8000 - initial_ram) / 64 

  'start and setup graphics
  gr.start
  gr.setup(COLOUR_MODE, X_TILES, Y_TILES, X_TILES*16>>COLOUR_MODE, Y_TILES*8, @offscreen, @map, initial_tiles, initial_ram)
  gr.db_setup(@onscreen, DOUBLE_BUFFER)

  vga.start(@vga_status)

  if RAM_RECYCLE
    'now that graphics cogs are loaded, reclaim the RAM that was used to hold their data
    gr.addram (gr.ramstart, gr.ramend)
    gr.addram (gr.db_ramstart, gr.db_ramend)
    gr.addram (vga.ramstart, vga.ramend)
    gr.addram (vga.lg_ramstart, vga.lg_ramend)

  'start sound effects manager
  sfx_flags := 0
  sfx_cnt_shield := 0
  shield_snd_ascending := true
  if SFX_COG
    cognew(SFX_manager, @sfx_stack)

  key.start(3)
  if RAM_RECYCLE
    'now that the key cog is loaded, reclaim the RAM that was used to hold its data
     gr.addram (key.ramstart, key.ramend)

  'start music (& sound effects) driver
  'hdmf.start(0)                ' sound effects and music
  hdmf.start(7)                 ' sound effects only
  if RAM_RECYCLE
    'now that the hdmf cog is loaded, reclaim the RAM that was used to hold its data
     gr.addram (hdmf.ramstart, hdmf.ramend)
  
  'start debugger
  '#IF LCD_DEBUGGER_ENABLE  
  { '<added by SCCuM preprocessor utility>
  debug.start
  debug.add_watch(@sfx_flags,debug#FORMAT__TYPE_HEX | debug#FORMAT__SIZE_8,string("sfx_flags"))  
  debug.add_watch(@curr_count,debug#FORMAT__TYPE_UINT,string("curr_count"))  
  debug.add_watch(@curr_count,debug#FORMAT__TYPE_INT,string("curr_count")) 
  debug.add_watch(@curr_count,debug#FORMAT__TYPE_HEX | debug#FORMAT__SIZE_32,string("curr_count"))
  debug.add_watch(@shot_time_to_live,debug#FORMAT__TYPE_HEX | debug#FORMAT__SIZE_16,string("time_live[0]"))
  debug.add_watch(@lives[0],debug#FORMAT__TYPE_HEX | debug#FORMAT__SIZE_8,string("lives[0]"))
  debug.add_watch(@object_dx[0],debug#FORMAT__TYPE_FIX,string("object_dx[0]"))
  debug.add_watch(@object_dy[0],debug#FORMAT__TYPE_FIX,string("object_dy[0]"))
  debug.add_watch(@object_dx[1],debug#FORMAT__TYPE_FIX,string("object_dx[1]"))
  debug.add_watch(@object_dy[1],debug#FORMAT__TYPE_FIX,string("object_dy[1]"))
  } '<added by SCCuM preprocessor utility>
  '#ENDIF


  '---------------------------------------------------------------
  'Show title screen
  {
  '#IF EEPROM_GRAPHICS_ENABLE
  waitcnt(cnt + 5000)
  gr.clear()
  'hdmf.EEPROM_Read(TILE_BUFFER, EEPROM_SCREEN_TITLE, CONSTANT((SCREEN_WIDTH * SCREEN_HEIGHT)>>3))  
  repeat until hdmf.EEPROM_IsDone
  gr.textmode(1,1,6,1)
  'gr.colorwidth(2,0)
  gr.colorwidth(1,0)
  gr.text(Constant(-SCREEN_WIDTH/2 + 1), constant(-SCREEN_HEIGHT/2 + 5), @version_string)
  if DOUBLE_BUFFER
    gr.copy_tpm(@onscreen, DOUBLE_BUFFER)

  'Wait for start button (or spacebar) to be pressed
  WaitForStartButton
  '#ENDIF
   }
  '---------------------------------------------------------------      
  
  '//// MAIN EXECUTION MANAGER
  repeat
    'PlaySong
    gr.colorwidth(1,0)
    if (gr.tiles > 120) & (COLOUR_MODE == 0)
      gr.textmode(7,12,6,5)
      gr.text(0, 150, @title1_string)
    elseif (gr.tiles > 100)
      gr.textmode(5,8,6,5)
      gr.text(0, 150, @title1_string)
    else
      gr.textmode(2,8,6,5)
      gr.text(0, 150, @title1_string)
    gr.textmode(1,1,6,5)
    gr.text(0, 53, @version_string)
    gr.text(0, -51, @title3_string)
    gr.text(0, -64, @title4_string) 
    gr.text(0, -111, @title_start_string) 
    'gr.colorwidth(3,3)
    'gr.box(-SCREEN_WIDTH/2+5,-SCREEN_HEIGHT/2+5,SCREEN_WIDTH-10,SCREEN_HEIGHT-10)
    if DOUBLE_BUFFER
      gr.copy(@onscreen)
    WaitForStartButton
    Configure_Options
    main_game

PUB main_game | i, j, k, base, base2, dx, dy, x, y, x2, y2, length, f, v, a, r, r_squared, player_index, shot_index, rock_index, watchval_int, watchval_fixed, died, thrust_snd_flag, shield_snd_flag                    
  
  ' initialize random variable
  random_var := cnt*171732
  random_var := ?random_var
  lives[PLAYER_0] := 5
  lives[PLAYER_1] := 5
  '#IF LCD_DEBUGGER_ENABLE
  { '<added by SCCuM preprocessor utility>
  debugger_key_held := false
  } '<added by SCCuM preprocessor utility>
  '#ENDIF

  'Initialize debug watch variables
  watchval_fixed := 0
  watchval_int   := gr.tiles

  '#IF MUSIC_ENABLE  
  'hdmf.play_song(EEPROM_SONG_FF1_BATTLE | hdmf#EEPROM_ASSET, hdmf#HDMF__LOOP)
  '#ENDIF
  

  ' generate random stars
  repeat i from 0 to NUM_STARS - 1
    star_x[i] := Rand_Range(CONSTANT (-SCREEN_WIDTH / 2), CONSTANT (SCREEN_WIDTH/2))
    star_y[i] := Rand_Range(CONSTANT (-SCREEN_HEIGHT / 2), CONSTANT (SCREEN_HEIGHT/2))


  repeat while (( lives[PLAYER_0] <> 0) & ( lives[PLAYER_1] <> 0))
  
    '//// Stop any ship action sounds in progress
    SFX_stop_ship_action_sounds(0)
    SFX_stop_ship_action_sounds(1)   
     
    '//// Initialize object info
    repeat i from 0 to NUM_OBJECTS - 1
      object_x[i] := 0        
      object_y[i] := 0        
      object_dx[i] := 0        
      object_dy[i] := 0
     
    '//// Initialize ship info
    object_x[OBJECT_PLAYER_0] := CONSTANT (((SCREEN_WIDTH / 2) - 30) << 16)
    object_y[OBJECT_PLAYER_0] := CONSTANT (((SCREEN_HEIGHT / 2) - 30) << 16)
    object_x[OBJECT_PLAYER_1] := CONSTANT (-(((SCREEN_WIDTH / 2) - 30) << 16))
    object_y[OBJECT_PLAYER_1] := CONSTANT (-(((SCREEN_HEIGHT / 2) - 30) << 16))
    ship_angle[OBJECT_PLAYER_0] := ANG_270
    ship_angle[OBJECT_PLAYER_1] := ANG_90
    shield_angle := ANG_0 
     
    repeat player_index from 0 to NUM_PLAYERS - 1
      ship_state[player_index]            := SHIP_STATE_ALIVE
      ship_flags[player_index]            := 0
      score[player_index]                 := 0
      ship_shield_power_fxp[player_index] := CONSTANT (STATUS_GAUGE_LENGTH << 16)    ' Fixed point
      ship_fuel_fxp[player_index]         := CONSTANT (STATUS_GAUGE_LENGTH << 16)    ' Fixed point
     
    '//// Initialize shot info
    repeat shot_index from 0 to NUM_SHOTS -1
      shot_time_to_live[shot_index] := 0

    '//// Initialize BFG
    bfg_flags := BFG_FLAG__OUT_OF_PLAY

    '//// Initialize black hole
    black_hole_particle_mask := 0

    '//// Initialize rocks
    repeat rock_index from 0 to NUM_ROCKS - 1
      rock_state[rock_index] := ROCK_STATE_ABSENT

    '//// Initialize explosion particles  
    repeat i from 0 to NUM_EXPLOSION_PARTICLES - 1
      'Create random fixed point dx's and dy's between -1.0 and 1.0
      'Note: The distrubution of these random points is currently a square.
      '      Would be better to pick points in polar space and make the distribution
      '      a circle, or even to force the density distribution so that it was
      '      more dense near the middle. 
      explosion_dx_fxp[i] := Rand_Range(-$0000ffff,$0000ffff)
      explosion_dy_fxp[i] := Rand_Range(-$0000ffff,$0000ffff) 
     
    ' BEGIN GAME LOOP ////////////////////////////////////////////////////
    died := 0
    
    damage[PLAYER_0] := 0
    damage[PLAYER_1] := 0

    gr.clear

    repeat while (died == 0)
      curr_count := cnt ' Save current counter value

      if not SFX_COG
        ' run the sounds effects periodically - not as good, but better than nothing!
        SFX_Manager

      shield_angle += CONSTANT($2000 / 31)
      
      '//// Read NES controller buttons
      if (game_options & OPTION_MONKEY_IS_ALIVE)
        'Simulate random input for testing
        if Rand_Range(1,20) =< 1
          nes_buttons := Rand_Range(0,$ffff)
        else
          nes_buttons := 0  
      else
        nes_buttons := NES_Read_Gamepad
      player_nes_buttons[0] := nes_buttons & $ff
      player_nes_buttons[1] := (nes_buttons >> 8)& $ff

      '//// Process keyboard keys
      
      if(key.keystate(119)) 'w'
        player_nes_buttons[PLAYER_1]|=CONSTANT(NES_UP)
      elseif(key.keystate(115)) 's'
        player_nes_buttons[PLAYER_1]|=CONSTANT(NES_DOWN)
      if(key.keystate(97)) 'a'
        player_nes_buttons[PLAYER_1]|=CONSTANT(NES_LEFT)
      elseif(key.keystate(100)) 'd'
        player_nes_buttons[PLAYER_1]|=CONSTANT(NES_RIGHT)
      if(key.keystate(121)) 'y'
        player_nes_buttons[PLAYER_1]|=CONSTANT(NES_A)
      elseif(key.keystate(103)) 'g'
        player_nes_buttons[PLAYER_1]|=CONSTANT(NES_B)

      '//// LCD Debugger control keys
      '#IF LCD_DEBUGGER_ENABLE
      { '<added by SCCuM preprocessor utility>
      'Process debugger keys
      if(key.keystate($c6)) 'Page UP'
        if not debugger_key_held
          debug.page_up
          debugger_key_held := TRUE
      elseif (key.keystate($c7)) 'Page DOWN'
        if not debugger_key_held   
          debug.page_down
          debugger_key_held := TRUE
      else
        debugger_key_held := FALSE
      } '<added by SCCuM preprocessor utility>
      '#ENDIF LCD_DEBUGGER_ENABLE
      {  
      '//// Debugging mode features
      if DEBUG__ENABLE_DEBUG_MODE
        ' Display debug mode indication
        gr.text(-20, 83, @debug_string)

        if(key.keystate(49)) '1'
          bfg_flags := BFG_FLAG__PICKUP
          bfg_x     := 115
          bfg_y     := 60
      }

      '//// Process both player ships
      repeat player_index from 0 to NUM_PLAYERS - 1
        'Determine sound flags for current plater
        if player_index == 0
           thrust_snd_flag := SFX_SND_SHIP0_THRUST
           shield_snd_flag := SFX_SND_SHIP0_SHIELD
        else
           thrust_snd_flag := SFX_SND_SHIP1_THRUST
           shield_snd_flag := SFX_SND_SHIP1_SHIELD

        if (ship_state[player_index] == SHIP_STATE_ALIVE) and not (ship_flags[player_index] & SHIP_FLAG_HYPERSPACE)
             
          '//// Process ROTATE
          button_temp := player_nes_buttons[player_index] ' This is a workaround for a spin compiler bug.
                                                          ' The compiler was not properly handling references
                                                          ' to player_nes_buttons[player_index] in the math below.    
          ' Process rotate left / right
          if((button_temp & NES_RIGHT) <> 0)
            if damage[player_index] => 3
              ' lots of damage
              ship_angle[player_index] := ship_angle[player_index] - CONSTANT (ANG_ROTATE << (SHL_ROTATE - 2))
            elseif damage[player_index] => 1
              ' a little damage
              ship_angle[player_index] := ship_angle[player_index] - CONSTANT (ANG_ROTATE << (SHL_ROTATE - 1))
            else
              ' no damage
              ship_angle[player_index] := ship_angle[player_index] - CONSTANT (ANG_ROTATE << SHL_ROTATE)
              
          if((button_temp & NES_LEFT) <> 0)
            if damage[player_index] => 3
              ' lots of damage
              ship_angle[player_index] := ship_angle[player_index] + CONSTANT (ANG_ROTATE << (SHL_ROTATE - 2))  
            elseif damage[player_index] => 1
              ' a little damage
              ship_angle[player_index] := ship_angle[player_index] + CONSTANT (ANG_ROTATE << (SHL_ROTATE - 1))  
            else
              ' no damage
              ship_angle[player_index] := ship_angle[player_index] + CONSTANT (ANG_ROTATE << SHL_ROTATE)  
           
          ' Bounds test ship angle
          if (ship_angle[player_index] > ANG_360)
            ship_angle[player_index] -= ANG_360
          elseif (ship_angle[player_index] < 0)
            ship_angle[player_index] +=ANG_360
         
          '//// Process HYPERSPACE
          if((button_temp & NES_DOWN) <> 0) and (game_options & OPTION_HYPERSPACE) 
            ship_flags[player_index] |= SHIP_FLAG_HYPERSPACE
            hyper_time[player_index] := SHIP_HYPERSPACE_TIME
         
          '//// Process SHIELDS
          if((button_temp & NES_UP) <> 0) and ship_shield_power_fxp[player_index] and (game_options & OPTION_SHIELDS) 
            ship_flags[player_index] |= SHIP_FLAG_SHIELDS_UP
         
            ' decrement shield power remaining
            if (ship_shield_power_fxp[player_index] > SHIELD_CONSUMPTION_RATE_FXP)
              ship_shield_power_fxp[player_index] -= SHIELD_CONSUMPTION_RATE_FXP
            else
              ship_shield_power_fxp[player_index] := 0
         
            'Play shield sound  
            sfx_flags |= shield_snd_flag     
            
          else
            ship_flags[player_index] &= !(SHIP_FLAG_SHIELDS_UP)
         
            'Stop shield sound
            sfx_flags &= !shield_snd_flag
         
          '//// Process FIRE
          if ((button_temp & NES_A) <> 0) & ((fire_debounce & (player_index + 1)) == 0)
            if (player_index == 0)
              i := 0
              j := CONSTANT (PLAYER_1_SHOT_START - 1)
            else
              i := CONSTANT (PLAYER_1_SHOT_START)
              j := CONSTANT (NUM_SHOTS - 1)
            repeat shot_index from i to j
              if (shot_time_to_live[shot_index] == 0)
                shot_time_to_live[shot_index] := SHOT_TTL_INIT + RAND_RANGE(-10,10)
                object_x[shot_index + OBJECT_PLAYER_0_SHOT_0] := object_x[player_index] + SinCos(COS, ship_angle[player_index]) * SHIP_GUN_PLACEMENT_DISTANCE
                object_y[shot_index + OBJECT_PLAYER_0_SHOT_0] := object_y[player_index] + SinCos(SIN, ship_angle[player_index]) * SHIP_GUN_PLACEMENT_DISTANCE 
                if (game_options & OPTION_INERTIAL_SHOTS)
                  ' Shot velocity is ship velocity + muzzle velocity
                  object_dx[shot_index + OBJECT_PLAYER_0_SHOT_0] := SCL_PROJECTILE_VEL*SinCos(COS, ship_angle[player_index]) + object_dx[player_index] 
                  object_dy[shot_index + OBJECT_PLAYER_0_SHOT_0] := SCL_PROJECTILE_VEL*SinCos(SIN, ship_angle[player_index]) + object_dy[player_index]
                else
                  ' Shot velocity is fixed regardless of ship velocity
                  object_dx[shot_index + OBJECT_PLAYER_0_SHOT_0] := SCL_PROJECTILE_VEL*SinCos(COS, ship_angle[player_index]) << 1
                  object_dy[shot_index + OBJECT_PLAYER_0_SHOT_0] := SCL_PROJECTILE_VEL*SinCos(SIN, ship_angle[player_index]) << 1
         
                fire_debounce |= (player_index + 1) ' Set the debounce flag
         
                'Play shot sound
                if (player_index == 0)   
                  sfx_flags |= SFX_SND_SHIP0_SHOT_START
                else
                  sfx_flags |= SFX_SND_SHIP1_SHOT_START 
                
                quit
          elseif ((button_temp & NES_A) == 0)
            fire_debounce &= !(player_index + 1) ' Clear the debounce flag
               
           
          '//// Process THRUST
          ' all calculations for ship thrust and position model are performed in fixed point math          
          if (player_nes_buttons[player_index] & NES_B) and (ship_fuel_fxp[player_index] or not(game_options & OPTION_FUEL))

            ' compute thrust vector, scale down cos/sin a bit to slow ship's acceleration
            if damage[player_index] => 3
              ' lots of damage
              thrust_dx[player_index] := SinCos(COS, ship_angle[player_index]) ~> (SAR_SHIP_VEL + 2)
              thrust_dy[player_index] := SinCos(SIN, ship_angle[player_index]) ~> (SAR_SHIP_VEL + 2)
            elseif damage[player_index] => 1
              ' a little damage
              thrust_dx[player_index] := SinCos(COS, ship_angle[player_index]) ~> (SAR_SHIP_VEL + 1)
              thrust_dy[player_index] := SinCos(SIN, ship_angle[player_index]) ~> (SAR_SHIP_VEL + 1)
            else
              ' no damage
              thrust_dx[player_index] := SinCos(COS, ship_angle[player_index]) ~> SAR_SHIP_VEL
              thrust_dy[player_index] := SinCos(SIN, ship_angle[player_index]) ~> SAR_SHIP_VEL
           
            ' apply thrust to ships current velocity
            object_dx[player_index] += thrust_dx[player_index]
            object_dy[player_index] += thrust_dy[player_index]
         
            ' decrement fuel remaining
            if ( ship_fuel_fxp[player_index] > FUEL_CONSUMPTION_RATE_FXP)
              ship_fuel_fxp[player_index] -= FUEL_CONSUMPTION_RATE_FXP
            else
              ship_fuel_fxp[player_index] := 0
         
            ' set thrusting flag
            ship_flags[player_index] |= SHIP_FLAG_THRUSTING
         
            ' play thrust sound
            sfx_flags |= thrust_snd_flag
            
          else
            ' clear thrusting flag
            ship_flags[player_index] &= !(SHIP_FLAG_THRUSTING)
         
            'stop thrust sound
            sfx_flags &= !thrust_snd_flag
            

          '//// Process Friction
          ' apply friction model, always in opposite direction of velocity
          ' frictional force is proportional to velocity, use power of 2 math to save time
          object_dx[player_index] -= object_dx[player_index] ~> SPACE_FRICTION
          object_dy[player_index] -= object_dy[player_index] ~> SPACE_FRICTION
          
        elseif not (ship_flags[player_index] & SHIP_FLAG_HYPERSPACE)
         
          '//// PROCESS DEGRADING SHIP STATE (EXPLODING -> DELAY -> DEAD) 
          if ship_state[player_index] > 0
            if (--ship_state[player_index] == 0)
              ' ship has gone thorough all the decaying states of death, now set 'died' flag so game round will end.
              died := 1
        else
          'ship must be in hyperspace
          if (hyper_time[player_index] <> 0)
            if (--hyper_time[player_index] == 0)
              ship_flags[player_index] &= !(SHIP_FLAG_HYPERSPACE)
              ' ship appears at random location - but not near edge (too hard to see!)
              object_x[player_index] := Rand_Range(CONSTANT (-SCREEN_WIDTH / 2 + 50), CONSTANT (SCREEN_WIDTH / 2 - 50))  << 16
              object_y[player_index] := Rand_Range(CONSTANT (-SCREEN_HEIGHT / 2 + 50), CONSTANT (SCREEN_HEIGHT / 2 - 50)) << 16
              object_dx[player_index] := 0
              object_dy[player_index] := 0
              ' always a chance of exploding when leaving hyperspace, chances increase with damage
              if Rand_Range(1,10) < damage[player_index] + 3
                --lives[player_index]
                ship_state[player_index] := SHIP_STATE_EXPLODING
                sfx_flags |= SFX_SND_EXPLODE
                SFX_stop_ship_action_sounds(player_index)   
                 

      repeat i from 0 to NUM_OBJECTS - 1
       
        '//// Process Black Hole
        '     NOTE: processed if enabled, and object is not a dead ship or a ship in hyperspace
        if (game_options & OPTION_BLACK_HOLE) and not ((i =< PLAYER_1) and ((ship_state[i] <> SHIP_STATE_ALIVE) or (ship_flags[i] & SHIP_FLAG_HYPERSPACE))) and not (((i => OBJECT_PLAYER_0_SHOT_0) and (i < OBJECT_ROCK_0)) and not (game_options & OPTION_SHOT_GRAVITY))
          ' now compute the acceleration toward the black hole, model based on F = (G*M1*M2)/r^2
          ' in other words, the force is equal to the product of the two masses times some constant G divided
          ' by the distance between the masses squared. Thus, we more or less need to accelerate the ship
          ' toward the black hole(s) proportional to some lumped constant divided by the distance to the black
          ' hole squared...
         
          ' compute each force direction vector d(dx ,dy) toward the black hole, so we can compute its length
          dx := BLACK_HOLE_X - (object_x[i] ~> 16)
          dy := BLACK_HOLE_Y - (object_y[i] ~> 16)
          r_squared := (dx*dx + dy*dy) ' no need to compute length r, when we are going to use r^2 in a moment 
         
          ' now compute the actual force itself, which is proportional to accel which in this sim will be used to change velocity each frame
          f := (BLACK_HOLE_GRAVITY / r_squared) <# GRAVITY_CLIP
                                                                        
          ' f can be thought of as acceleration since its proportional to mass which is virtual and can be assumed to be 1
          ' thus we can use it to create a velocity vector toward the black hole now in the direction of of the vector d(dx, dy)
          dx := dx << 16 ' convert to fixed point
          dy := dy << 16
         
          ' compute length of d which is just r, careful to compute fixed point values properly
          r := (^^(r_squared << 16)) >> 8 ' square root operation turns 16.16 into 24.8. Convert r to integer
         
          ' normalize the vector and scale by force magnityde
          dx := (f~>8) * ((dx / r) ~> 8)
          dy := (f~>8) * ((dy / r) ~> 8)
         
          ' update velocity with acceleration due to black hole
          if (game_options & OPTION_NEGATIVE)
            object_dx[i] -= dx
            object_dy[i] -= dy
          else
            object_dx[i] += dx
            object_dy[i] += dy
         
          'if object is a player ship
          if (i < OBJECT_PLAYER_0_SHOT_0)
            'Check for Collision with black hole
            if ship_state[i] == SHIP_STATE_ALIVE
              if (object_x[i] > CONSTANT (BLACK_HOLE_X - COLLISION_RANGE_FP)) & (object_x[i] < CONSTANT (BLACK_HOLE_X + COLLISION_RANGE_FP))
                if (object_y[i] > CONSTANT (BLACK_HOLE_Y - COLLISION_RANGE_FP)) & (object_y[i] < CONSTANT (BLACK_HOLE_Y + COLLISION_RANGE_FP))
                  --lives[i]
                  ship_state[i] := SHIP_STATE_EXPLODING
                  sfx_flags |= SFX_SND_EXPLODE
                  'Colliding with the black hole tends to fling your exploding bits super fast across
                  '  space, which looks silly, so chop the velocity down
                  object_dx[i] ~>= 2
                  object_dy[i] ~>= 2
                  SFX_stop_ship_action_sounds(i)   
         
        ' CLAMP OBJECT VELOCITIES ///////////////////////////////////////
        ' clamp maximum velocity, otherwise ship will get going light speed due to black holes when the distance approaches 0!
        dx := (object_dx[i] ~> 16)
        dy := (object_dy[i] ~> 16)           
         
        ' test if object velocity greater than threshold
        v := ^^(dx*dx + dy*dy)
         
        ' perform comparison (to squared max, to make math easier)
        if (i < OBJECT_PLAYER_0_SHOT_0)
          ' Object is a ship
          if (v > MAX_SHIP_VEL)
            ' scale velocity vector back approx 1/8th
            object_dx[i] := MAX_SHIP_VEL*(object_dx[i] / v)                           
            object_dy[i] := MAX_SHIP_VEL*(object_dy[i] / v)
        else
          ' Object is a projectile (or a rock)
          if (v > MAX_PROJECTILE_VEL)
            ' scale velocity vector back approx 1/8th
            object_dx[i] := MAX_PROJECTILE_VEL*(object_dx[i] / v)                     
            object_dy[i] := MAX_PROJECTILE_VEL*(object_dy[i] / v)
         
         
        ' UPDATE OBJECT POSITIONS ///////////////////////////////////////
        ' Note: Time is currently wasted updating "inactive" shots, in order to
        ' conserve the code space necessary to test for those cases.1
        object_x[i] += object_dx[i]                    
        object_y[i] += object_dy[i]
         
        ' screen bounds test for objects
        if (game_options & OPTION_BOUNCE)
          'bounce back - bounce off edge
          if (object_x[i] > CONSTANT ((SCREEN_WIDTH/2)<< 16))
            'object_x[i] := CONSTANT ((SCREEN_WIDTH/2 - 1)<< 16)
            object_dx[i] := -object_dx[i]
            if i < NUM_PLAYERS
              ship_angle[i] := ANG_180 - ship_angle[i]
          elseif (object_x[i] < CONSTANT ((-SCREEN_WIDTH/2)<<16))
            'object_x[i] := CONSTANT ((-SCREEN_WIDTH/2 + 1)<< 16)
            object_dx[i] := -object_dx[i]
            if i < NUM_PLAYERS
              ship_angle[i] := ANG_180 - ship_angle[i]
          if (object_y[i] > CONSTANT ((SCREEN_HEIGHT/2)<<16))
            'object_y[i] := CONSTANT ((SCREEN_HEIGHT/2 - 1)<< 16)
            object_dy[i] := -object_dy[i] 
            if i < NUM_PLAYERS
              ship_angle[i] := - ship_angle[i]
          elseif (object_y[i] < CONSTANT ((-SCREEN_HEIGHT/2)<<16))
            'object_y[i] := CONSTANT ((-SCREEN_HEIGHT/2 + 1)<< 16)
            object_dy[i] := -object_dy[i]
            if i < NUM_PLAYERS
              ship_angle[i] := - ship_angle[i] 

          if i < NUM_PLAYERS
          ' Bounds test ship angle
            if (ship_angle[i] > ANG_360)
              ship_angle[i] -= ANG_360
            elseif (ship_angle[i] < ANG_0)
              ship_angle[i] +=ANG_360

        elseif (game_options & OPTION_EXPANDED)

          'explanded universe - return on other side but universe is 1/4 bigger than screen on every side
          if (object_x[i] > CONSTANT ((3*SCREEN_WIDTH/4)<< 16))
            object_x[i] -= CONSTANT ((3*SCREEN_WIDTH/2)<<16)
          elseif (object_x[i] < CONSTANT ((-3*SCREEN_WIDTH/4)<<16))
            object_x[i] += CONSTANT ((3*SCREEN_WIDTH/2)<<16)
          if (object_y[i] > CONSTANT ((3*SCREEN_HEIGHT/4)<<16))
            object_y[i] -= CONSTANT ((3*SCREEN_HEIGHT/2)<<16)
          elseif (object_y[i] < CONSTANT ((-3*SCREEN_HEIGHT/4)<<16))
            object_y[i] += CONSTANT ((3*SCREEN_HEIGHT/2)<<16)
        
        else
          'open universe - return on other side
          if (object_x[i] > CONSTANT ((SCREEN_WIDTH/2)<< 16))
            object_x[i] -= CONSTANT (SCREEN_WIDTH<<16)
          elseif (object_x[i] < CONSTANT ((-SCREEN_WIDTH/2)<<16))
            object_x[i] += CONSTANT (SCREEN_WIDTH<<16)
          if (object_y[i] > CONSTANT ((SCREEN_HEIGHT/2)<<16))
            object_y[i] -= CONSTANT (SCREEN_HEIGHT<<16)
          elseif (object_y[i] < CONSTANT ((-SCREEN_HEIGHT/2)<<16))
            object_y[i] += CONSTANT (SCREEN_HEIGHT<<16)
        
      ' PLAYER SHOTS //////////////////////////////////////////////////
      repeat shot_index from 0 to NUM_SHOTS -1
        if (shot_time_to_live[shot_index] <> 0)
          --shot_time_to_live[shot_index]
          ' shot to ship
          repeat player_index from PLAYER_0 to PLAYER_1
          
            ' ignore ships in hyperspace
            if (ship_flags & SHIP_FLAG_HYPERSPACE) <> 0
              next
          'if (shot_index < PLAYER_1_SHOT_START)
        '    player_index := PLAYER_1
          'else
          '  player_index := PLAYER_0
         
            '//// SHOT TO SHIELD COLLISIONS
         
            ' If the target ship's shields are up
            if ship_flags[player_index] & SHIP_FLAG_SHIELDS_UP
              dx := (object_x[shot_index+OBJECT_PLAYER_0_SHOT_0] - object_x[player_index]) ~> 16   ' dx, as integer
              dy := (object_y[shot_index+OBJECT_PLAYER_0_SHOT_0] - object_y[player_index]) ~> 16   ' dy, as integer
              r_squared := (dx*dx + dy*dy)
              ' If the shot is within the shield radius
              if r_squared < CONSTANT(SHIELD_RADIUS * SHIELD_RADIUS)
                'if true
                  'SIMPLE REVERSE
                  object_dx[shot_index+OBJECT_PLAYER_0_SHOT_0] := -object_dx[shot_index+OBJECT_PLAYER_0_SHOT_0]  + RAND_RANGE(-10,10) << 16
                  object_dy[shot_index+OBJECT_PLAYER_0_SHOT_0] := -object_dy[shot_index+OBJECT_PLAYER_0_SHOT_0]  + RAND_RANGE(-10,10) << 16
                'else
                  'COMPLEX REVERSE
         
            '//// SHOT TO SHIP COLLISIONS
            ShotToShip(shot_index, player_index)
         
          '//// SHOT TO ROCK COLLISIONS
          if game_options & OPTION_ROCKS
            repeat rock_index from 0 to NUM_ROCKS - 1
               ShotToRock(shot_index, rock_index)
                  
      ' PLAYER COLLISION /////////////////////////////////////////////////////
      if (ship_state[PLAYER_0] == SHIP_STATE_ALIVE) and (ship_state[PLAYER_1] == SHIP_STATE_ALIVE) and (((ship_flags[PLAYER_0] | ship_flags[PLAYER_1]) & SHIP_FLAG_HYPERSPACE) == 0)
        'NOTE: Ship/Ship collision detection uses 2x the standard Shot/Ship collision detection range
        if (object_x[PLAYER_0] > (object_x[PLAYER_1] - CONSTANT(COLLISION_RANGE_FP<<1))) & (object_x[PLAYER_0] < (object_x[PLAYER_1] + CONSTANT(COLLISION_RANGE_FP<<1)))
          if (object_y[PLAYER_0] > (object_y[PLAYER_1] - CONSTANT(COLLISION_RANGE_FP<<1))) & (object_y[PLAYER_0] < (object_y[PLAYER_1] + CONSTANT(COLLISION_RANGE_FP<<1)))
            if not (ship_flags[PLAYER_0] & SHIP_FLAG_SHIELDS_UP)
              KillPlayer(PLAYER_0)
            if not (ship_flags[PLAYER_1] & SHIP_FLAG_SHIELDS_UP)
              KillPlayer(PLAYER_1)

      ' ROCK COLLISION /////////////////////////////////////////////////////
      if game_options & OPTION_ROCKS
        repeat rock_index from 0 to NUM_ROCKS - 1
          repeat player_index from 0 to NUM_PLAYERS - 1
            if (ship_state[player_index] == SHIP_STATE_ALIVE) and (rock_state[rock_index] == ROCK_STATE_PRESENT) and ((ship_flags[player_index] & SHIP_FLAG_HYPERSPACE) == 0)
              'NOTE: Rock to Rock collision detection uses 2x the standard Shot/Ship collision detection range
              if (object_x[player_index] > (object_x[rock_index+OBJECT_ROCK_0] - CONSTANT(COLLISION_RANGE_FP<<1))) & (object_x[player_index] < (object_x[rock_index+OBJECT_ROCK_0] + CONSTANT(COLLISION_RANGE_FP<<1)))
                if (object_y[player_index] > (object_y[rock_index+OBJECT_ROCK_0] - CONSTANT(COLLISION_RANGE_FP<<1))) & (object_y[player_index] < (object_y[rock_index+OBJECT_ROCK_0] + CONSTANT(COLLISION_RANGE_FP<<1)))
                  KillPlayer(player_index)

      '==================================================================================================
      ' RENDERING SECTION 
      '==================================================================================================
      
      '//// INITIALE TEXT COLOR AND SIZE
      gr.textmode(1,1,5,3)
      gr.colorwidth(1,0)
      
      '//// DRAW DEBUGGING WATCH VARIABLES
      if DEBUG__SHOW_WATCH_VARIABLES
        'Draw fixed point watch variable
        Fixed_To_String(@fixed_pt_string_0, watchval_fixed )
        gr.text(PLAYER_1_STATS_X_POS, PLAYER_1_STATS_Y_POS - 20, @fixed_pt_string_0)
         
        'Draw integer watch variable
        Signed_Int_To_String(@int_string_0, watchval_int )
        gr.text(PLAYER_0_STATS_X_POS, PLAYER_0_STATS_Y_POS - 22, @int_string_0)
      

      '//// DRAW STARS
      if (game_options & OPTION_TWINKLE)
        'twinkle stars randomly
        repeat i from 0 to NUM_STARS - 1
          if (rand_range(0,10) > COLOUR_MODE)
            gr.colorwidth(1,0)
            gr.plot(star_x[i], star_y[i])
          elseif COLOUR_MODE > 0
            gr.colorwidth(2,0)
            gr.plot(star_x[i], star_y[i])
      else
        ' simple plot
        repeat i from 0 to NUM_STARS - 1
          gr.plot(star_x[i], star_y[i])

      '//// DRAW ROCKS
      if (game_options & OPTION_ROCKS)
         'draw rocks that are visible, or make new ones appear at random, up to NUM_ROCKS
        repeat i from 0 to NUM_ROCKS - 1
          if (rock_state[i] == ROCK_STATE_PRESENT)
          ' rock is visible, so spin it and then draw it
            rock_angle[i] += rock_spin[i]
            if (rock_angle[i] > ANG_360)
              rock_angle[i] -= ANG_360
            elseif (rock_angle[i] < ANG_0)
              rock_angle[i] +=ANG_360
            gr.colorwidth(1,0)
            gr.vec(object_x[i+OBJECT_ROCK_0]~>16, object_y[i+OBJECT_ROCK_0]~>16, rock_scale[i], rock_angle[i], @vectors_rock) 

          elseif (rock_state[i] =< ROCK_STATE_EXPLODING)
            if (rock_state[i] > ROCK_STATE_DIE_DELAY)
              'Rock is in the ROCK_STATE_EXPLODING state, so draw explosion
              x := object_x[i+OBJECT_ROCK_0]~>16
              y := object_y[i+OBJECT_ROCK_0]~>16
              'Set j to a fixed point value between 0.0 and 1.0 reflecting the progress of the
              '   player explosion state.  Starts at 0.0 at beginning of explosion, and
              '   ends at 1.0 at the end of the explosion.
              j := $00010000 - (((rock_state[i] - ROCK_STATE_DIE_DELAY) << 16) / CONSTANT(ROCK_STATE_EXPLODING - ROCK_STATE_DIE_DELAY))
              'Scale j by 90 degrees and convert back to integer.  Now j will range from 0 to 90 degrees
              '   over the course of the explosion.
              j := (j * ANG_90) >> 16
              'Compute a fixed point value which ranges from 0.0 to the explosion radius over the course 
              '   of the explosion in a sinusoidal expansion (rapidly at first, then slowly at the end).
              '   This value will be used to scale the displayed distance of the explosion particles
              '   from the ship's center.
              j := SinCos(SIN, j) * EXPLOSION_RADIUS 
              'Loop over all the explosion particles 
              repeat k from 0 to CONSTANT (NUM_EXPLOSION_PARTICLES - 1)
                'Draw an explosion particle 
                gr.colorwidth(rand_range(1,3),1)
                gr.plot((explosion_dx_fxp[k] * j~>16)~>16 + x, (explosion_dy_fxp[k] * j~>16)~>16 + y)
            if rock_state[i] > ROCK_STATE_ABSENT
              '//// PROCESS DEGRADING ROCK STATE (EXPLODING -> DELAY -> ABSENT)
              --rock_state[i]
            elseif rand_range(0, 3000) =< 1
              ' rocks make an appearance at random times
              rock_state[i] := ROCK_STATE_PRESENT
              rock_angle[i] := ANG_1 * rand_range(1,359)
              rock_scale[i] := CONSTANT (SHIP_SCALE/2) + rand_range(0, CONSTANT (SHIP_SCALE/2))
              rock_spin[i] := ANG_1 * rand_range(-5, 5)
              ' rock appears on a random edge - but not right near ships!
              if rand_range(0,1) == 0 
                object_x[i+OBJECT_ROCK_0] := CONSTANT ((SCREEN_WIDTH/2 - 1)<<16)
                object_y[i+OBJECT_ROCK_0] := rand_range (CONSTANT (-SCREEN_HEIGHT/2 + 50), CONSTANT (SCREEN_HEIGHT/2 - 50))<<16
              else
                object_x[i+OBJECT_ROCK_0] := rand_range (CONSTANT (-SCREEN_WIDTH/2 + 100), CONSTANT (SCREEN_WIDTH/2 - 100))<<16
                object_y[i+OBJECT_ROCK_0] := CONSTANT ((SCREEN_HEIGHT/2 - 1)<<16)
              object_dx[i+OBJECT_ROCK_0] := rand_range (-2,2)<<16
              object_dy[i+OBJECT_ROCK_0] := rand_range (-2,2)<<16
              if (object_dx[i+OBJECT_ROCK_0] == 0) and (object_dy[i+OBJECT_ROCK_0] == 0)
                ' don't let initial speed be zero
                object_dx[i+OBJECT_ROCK_0] := 1<<16
            
      '//// DRAW PLAYER SHIPS AND STATS
      repeat player_index from 0 to NUM_PLAYERS - 1
        
        if not (ship_flags[player_index] & SHIP_FLAG_HYPERSPACE)
          '////---------------------
          '//// DRAW SHIPS
          '////---------------------
          ' extract whole parts of fixed point position to save time in parameters
          x := object_x[player_index] ~> 16
          y := object_y[player_index] ~> 16
           
          if (ship_state[player_index] =< SHIP_STATE_EXPLODING) and (ship_state[player_index] > SHIP_STATE_DIE_DELAY)
            'Ship is in the SHIP_STATE_EXPLODING state, so draw explosion

            if ship_state[player_index] > SHIP_STATE_STRICKEN
              if (player_index == PLAYER_0)
                gr.colorwidth(2,1)
                gr.vec(x, y, SHIP_SCALE, ship_angle[player_index], player0_ship[damage[player_index]]) 
              else
                gr.colorwidth(3,1)
                gr.vec(x, y, SHIP_SCALE, ship_angle[player_index], player1_ship[damage[player_index]])
            else
              'Set j to a fixed point value between 0.0 and 1.0 reflecting the progress of the
              '   player explosion state.  Starts at 0.0 at beginning of explosion, and
              '   ends at 1.0 at the end of the explosion.
              j := $00010000 - (((ship_state[player_index] - SHIP_STATE_DIE_DELAY) << 16) / CONSTANT(SHIP_STATE_EXPLODING - SHIP_STATE_DIE_DELAY))
              'Scale j by 90 degrees and convert back to integer.  Now j will range from 0 to 90 degrees
              '   over the course of the explosion.
              j := (j * ANG_90) >> 16
              'Compute a fixed point value which ranges from 0.0 to the explosion radius over the course 
              '   of the explosion in a sinusoidal expansion (rapidly at first, then slowly at the end).
              '   This value will be used to scale the displayed distance of the explosion particles
              '   from the ship's center.
              j := SinCos(SIN, j) * EXPLOSION_RADIUS 
              'Loop over all the explosion particles 
              repeat i from 0 to CONSTANT (NUM_EXPLOSION_PARTICLES - 1)
                'Draw an explosion particle 
                gr.colorwidth(rand_range(1,3),1)
                gr.plot((explosion_dx_fxp[i] * j~>16)~>16 + x, (explosion_dy_fxp[i] * j~>16)~>16 + y)
            
          else
            if (ship_state[player_index] == SHIP_STATE_ALIVE)
              'Ship is alive, so draw it
              if (player_index == PLAYER_0)
                gr.colorwidth(2,0)
                gr.vec(x, y, SHIP_SCALE, ship_angle[player_index], player0_ship[damage[player_index]]) 
              else
                gr.colorwidth(3,0)
                gr.vec(x, y, SHIP_SCALE, ship_angle[player_index], player1_ship[damage[player_index]])
                
              gr.colorwidth(1,0)
              
              'If shields are up, then show them
              if (ship_flags[player_index] & SHIP_FLAG_SHIELDS_UP)  
                gr.vec(x, y, SHIP_SCALE, shield_angle, @vectors_shield)
           
              'Draw thruster
              if ((ship_flags[player_index] & SHIP_FLAG_THRUSTING ) and (?random_var & $01))
                'Offset the flare from ship center 
                if player_index
                  'Player 1 
                  'gr.colorwidth(1,0)
                  dx := x - (SinCos(COS, ship_angle[PLAYER_1]) * P1_THRUSTER_DISTANCE) ~> 16
                  dy := y - (SinCos(SIN, ship_angle[PLAYER_1]) * P1_THRUSTER_DISTANCE) ~> 16     
                else
                  'Player 0
                  'gr.colorwidth(2,0)
                  dx := x - (SinCos(COS, ship_angle[PLAYER_0]) * P0_THRUSTER_DISTANCE) ~> 16
                  dy := y - (SinCos(SIN, ship_angle[PLAYER_0]) * P0_THRUSTER_DISTANCE) ~> 16
                gr.plot(dx, dy)
                gr.line(dx - (thrust_dx[player_index] ~> ROR_THRUST), dy - (thrust_dy[player_index] ~> ROR_THRUST) )
                                                                                           
        '////---------------------
        '//// DRAW DAMAGE
        '////---------------------
        repeat i from 0 to CONSTANT (4 * NUM_DAMAGE_PARTICLES - 1)
          if damage_ttl[i] > 0
            damage_ttl[i]--
            if i < 2*NUM_DAMAGE_PARTICLES
              gr.colorwidth(2,1)
            else
              gr.colorwidth(3,1)
            damage_x[i] += damage_dx[i]
            damage_y[i] += damage_dy[i]
            gr.plot(damage_x[i],damage_y[i])
            
        '////---------------------
        '//// DRAW PLAYER STATS 
        '////---------------------
        gr.colorwidth(1,0)
        if player_index == 0
          x := PLAYER_0_STATS_X_POS
          y := PLAYER_0_STATS_Y_POS
        else
          x := PLAYER_1_STATS_X_POS
          y := PLAYER_1_STATS_Y_POS
      
        '//// DRAW FUEL REMAINING        
        if ship_fuel_fxp[player_index] and (game_options & OPTION_FUEL)
          gr.plot(x                                  , y + 7) 
          gr.line(x + ship_fuel_fxp[player_index]>>16, y + 7)
          gr.plot(x                                  , y + 8) 
          gr.line(x + ship_fuel_fxp[player_index]>>16, y + 8)          
      
        '//// DRAWSHIELD POWER REMAINING
        if ship_shield_power_fxp[player_index] and (game_options & OPTION_SHIELDS)
          j := ship_shield_power_fxp[player_index] >> 16
          repeat i from x to x + j step 2                            
            gr.plot(i, y + 10)
            gr.plot(i, y + 11)  
      
      '//// DRAW SHOTS
      'gr.colorwidth(1,1)
      gr.colorwidth(1,0)
      repeat shot_index from 0 to NUM_SHOTS - 1
        if (shot_time_to_live[shot_index] <> 0)
          ' Draw 2x2 shot
          x := object_x[shot_index+OBJECT_PLAYER_0_SHOT_0] ~> 16
          y := object_y[shot_index+OBJECT_PLAYER_0_SHOT_0] ~> 16
          gr.plot(x  ,y  )
          gr.plot(x+1,y+1)
          gr.plot(x+1,y  )
          gr.plot(x  ,y+1)
          
      '//// DRAW SHIPS REMAINING
      gr.colorwidth(1,0)
      if (lives[PLAYER_0] <> 0)
        'gr.colorwidth(2,0)
        repeat i from 1 to lives[PLAYER_0]
          gr.vec(PLAYER_0_STATS_X_POS + (i - 1) << 3 + 4, PLAYER_0_STATS_Y_POS, $0015, CONSTANT ($2000 >> 2), player0_ship[0])
      if (lives[PLAYER_1] <> 0)
        'gr.colorwidth(3,0)
        repeat i from 1 to lives[PLAYER_1]
          gr.vec(PLAYER_1_STATS_X_POS + (i - 1) << 3 + 4, PLAYER_1_STATS_Y_POS, $0015, CONSTANT ($2000 >> 2), player1_ship[0])

      '//// DRAW BFG
      'if bfg_flags & BFG_FLAG__PICKUP
      '    gr.vec(bfg_x, bfg_y, $0020, shield_angle, @vectors_bfg)

      '//// PROCESS AND RENDER BLACK HOLE
      if ((game_options & OPTION_BLACK_HOLE) <> 0)
        {
        '//// Old style black hole animation
        gr.colorwidth(1,0)
        length := 8
        repeat 5
          gr.plot(BLACK_HOLE_X,BLACK_HOLE_Y)
          gr.line(BLACK_HOLE_X + Rand_Range(-length,length), BLACK_HOLE_Y + Rand_Range(-length,length) )
        }
        gr.colorwidth(1,0)
        gr.plot(0,0)
        j := 1
        repeat i from 0 to CONSTANT (NUM_BLACK_HOLE_PARTICLES - 1)
          if not(black_hole_particle_mask & j)
             'Particle does not exist.  Randomly choose whether to spawin it
             if Rand_Range(1,20) < 2
               'Spawn particle
               black_hole_particle_angle[i] := Rand_Range(ANG_0, ANG_360)
               black_hole_particle_radius[i] := CONSTANT (20 << 16)
               black_hole_particle_mask |= j
          else
             'Particle exists so render and update it
             gr.plot((SinCos(COS, black_hole_particle_angle[i]) * (black_hole_particle_radius[i] >> 16))~>16,  (SinCos(SIN, black_hole_particle_angle[i]) * (black_hole_particle_radius[i]>>16))~>16)
             'gr.plot((SinCos(COS, black_hole_particle_angle[i]) * 6)~>16,  (SinCos(SIN, black_hole_particle_angle[i]) * 6)~>16) 
             black_hole_particle_radius[i] -= $8000
             black_hole_particle_angle[i] += CONSTANT(ANG_1 * 3) * (9 - (black_hole_particle_radius[i]>>16))
             if(black_hole_particle_angle[i] > ANG_360)
               black_hole_particle_angle[i] -= ANG_360
               
             'If particle has reached the center
             if  black_hole_particle_radius[i] < 0
                'Extinguish particle
                black_hole_particle_mask &= !j
          j <<= 1


      
      '==================================================================================================
      ' PACING
      '==================================================================================================
      {
      'OLD Pacing algorithm
      if (curr_count>>1 < $7fffffff - CONSTANT(PACING_TICK>>1) )
        'No counter wrap
        if (curr_count>>1) + CONSTANT(PACING_TICK>>1 - PACING_MARGIN>>1) > (cnt>>1)
          'Wait for pacing
          waitcnt(curr_count + PACING_TICK)
        else
          if DEBUG__SHOW_PACING_OVERRUNS
            'Missed target frame rate.  Indicate with a large square dot in the corner, and continue on
            gr.plot(CONSTANT(SCREEN_WIDTH/2-1), CONSTANT(SCREEN_HEIGHT/2-1))
            gr.plot(CONSTANT(SCREEN_WIDTH/2-2), CONSTANT(SCREEN_HEIGHT/2-1))
            gr.plot(CONSTANT(SCREEN_WIDTH/2-1), CONSTANT(SCREEN_HEIGHT/2-2))
            gr.plot(CONSTANT(SCREEN_WIDTH/2-2), CONSTANT(SCREEN_HEIGHT/2-2))
      else
        if DEBUG__SHOW_PACING_OVERRUNS
          'Counter wrap
          gr.plot(-CONSTANT(SCREEN_WIDTH/2-1), CONSTANT(SCREEN_HEIGHT/2-1))
          gr.plot(-CONSTANT(SCREEN_WIDTH/2-2), CONSTANT(SCREEN_HEIGHT/2-1))
          gr.plot(-CONSTANT(SCREEN_WIDTH/2-1), CONSTANT(SCREEN_HEIGHT/2-2))
          gr.plot(-CONSTANT(SCREEN_WIDTH/2-2), CONSTANT(SCREEN_HEIGHT/2-2))
      }

      {
      'NEW Pacing algorithm (DEBUG__SHOW_PACING_OVERRUNS not currently supported)
      if DEBUG__SHOW_PACING_LED
        OUTA [0] := 1 'Set debug LED  
      
      pacing_target := (curr_count + PACING_TICK) >> 2
      pacing_cnt := cnt >> 2
      if(pacing_target < CONSTANT($7fffffff >> 2))
        'No wrap
        j := pacing_target + CONSTANT($7fffffff >> 2)
        repeat while(pacing_cnt < pacing_target) or (pacing_cnt > j)
          pacing_cnt := cnt >> 2
      else
        'Wrap
        j := pacing_target - CONSTANT($7fffffff >> 2)
        repeat while(pacing_cnt < pacing_target) and (pacing_cnt > j)
          pacing_cnt := cnt >> 2

      'Clear debug LED
      OUTA [0] := 0 'Clear debug LED
      }

      'move bitmap to display if double buffering
      if DOUBLE_BUFFER
        gr.move(@onscreen)
      else
        gr.clear
     
    ' END MAIN GAME LOOP REPEAT BLOCK //////////////////////////////////

  '----------------------------
  ' Game Over
  '----------------------------
  
  '#IF MUSIC_ENABLE
  'hdmf.stop_song  'stop playing music
  '#ENDIF
  
  SFX_stop_ship_action_sounds(0)
  SFX_stop_ship_action_sounds(1)
  
  gr.textmode(4,1,5,3)
  gr.colorwidth(1,0)
  gr.text(-90, 30, @game_over_string)
  if DOUBLE_BUFFER
    gr.copy(@onscreen)
  waitcnt(70_000_000 + cnt)
  
  '#IF MUSIC_ENABLE
  'hdmf.play_song(EEPROM_SONG_LOST_LOVE | hdmf#EEPROM_ASSET, hdmf#HDMF__LOOP)
  'hdmf.play_song(EEPROM_SONG_GENTLE_WIND | hdmf#EEPROM_ASSET, hdmf#HDMF__LOOP)
  'hdmf.play_song(EEPROM_SONG_NORTH_WALL | hdmf#EEPROM_ASSET, hdmf#HDMF__LOOP) 
  '#ENDIF
  
  gr.textmode(1,1,5,3) 
  gr.text(-34, -35, @game_over_restart)
  if DOUBLE_BUFFER
    gr.copy(@onscreen)
  
  '//// WAIT FOR RESTART
  WaitForStartButton
  gr.clear
  if DOUBLE_BUFFER
    gr.copy(@onscreen)

  '#IF MUSIC_ENABLE
  'hdmf.stop_song  'stop playing music
  '#ENDIF
    
' ////////////////////////////////////////////////////////////////////

PRI KillRock(rock_index)
  if rock_state[rock_index] == ROCK_STATE_PRESENT
    rock_state[rock_index] := ROCK_STATE_EXPLODING
  sfx_flags |= SFX_SND_EXPLODE   

PRI KillPlayer(player_index)
  if lives[player_index] > 0
    --lives[player_index]
  ship_state[player_index] := SHIP_STATE_EXPLODING
  sfx_flags |= SFX_SND_EXPLODE   
  SFX_stop_ship_action_sounds(player_index)    

PRI ShotDamage(player_index, shot_index) | new_damage, i, j
  'only process live shots
  if shot_time_to_live[shot_index] > 0
    if (damage[player_index] => 3) or (shot_time_to_live[shot_index] > CONSTANT (SHOT_TTL_INIT*3/4)) or not (game_options & OPTION_DAMAGE)
      'definite kill if already badly damaged, or if short range
      KillPlayer(player_index)
    else
      'just more damage 
      if (damage[player_index] == 0)
        new_damage := rand_range(0,1)
        damage[player_index] := new_damage + 1
      else
        new_damage := 2 - damage[player_index]
        damage[player_index] := 3
       'show some debris
      repeat i from 0 to CONSTANT (NUM_DAMAGE_PARTICLES - 1)
        j := NUM_DAMAGE_PARTICLES * (player_index * 2 + new_damage) + i
        damage_x[j] := object_x[player_index] ~> 16   
        damage_y[j] := object_y[player_index] ~> 16   
        damage_dx[j] := rand_range(-2,2)    
        damage_dy[j] := rand_range(-2,2)
        damage_ttl[j] := DAMAGE_TTL_INIT    
      sfx_flags |= SFX_SND_DAMAGE
    SFX_stop_ship_action_sounds(player_index)

PRI ShotToShip(shot_index, player_index)
  if (ship_state[player_index] == SHIP_STATE_ALIVE) and not (ship_flags[player_index] & SHIP_FLAG_SHIELDS_UP)  
    if (object_x[shot_index+OBJECT_PLAYER_0_SHOT_0] > (object_x[player_index] - COLLISION_RANGE_FP)) & (object_x[shot_index+OBJECT_PLAYER_0_SHOT_0] < (object_x[player_index] + COLLISION_RANGE_FP))
      if (object_y[shot_index+OBJECT_PLAYER_0_SHOT_0] > (object_y[player_index] - COLLISION_RANGE_FP)) & (object_y[shot_index+OBJECT_PLAYER_0_SHOT_0] < (object_y[player_index] + COLLISION_RANGE_FP))
        'kill or damage - depends on range
        ShotDamage(player_index,shot_index)
       'absorb the shot
        shot_time_to_live[shot_index] := 0
        
PRI ShotToRock(shot_index, rock_index)
  if (rock_state[rock_index] == ROCK_STATE_PRESENT)  
    if (object_x[shot_index+OBJECT_PLAYER_0_SHOT_0] > (object_x[rock_index+OBJECT_ROCK_0] - COLLISION_RANGE_FP)) & (object_x[shot_index+OBJECT_PLAYER_0_SHOT_0] < (object_x[rock_index+OBJECT_ROCK_0] + COLLISION_RANGE_FP))
      if (object_y[shot_index+OBJECT_PLAYER_0_SHOT_0] > (object_y[rock_index+OBJECT_ROCK_0] - COLLISION_RANGE_FP)) & (object_y[shot_index+OBJECT_PLAYER_0_SHOT_0] < (object_y[rock_index+OBJECT_ROCK_0] + COLLISION_RANGE_FP))
        KillRock(rock_index)
       'absorb the shot
        shot_time_to_live[shot_index] := 0

PUB WaitForStartButton
  nes_buttons := 0
  'hang here until 'START' or '<SPACEBAR>' is pressed    
  repeat while not (nes_buttons & CONSTANT( NES_START | (NES_START << 8) ))  and not (key.keystate(32)) and not((game_options & OPTION_MONKEY_IS_ALIVE) and (Rand_Range(1,100) == 1)) 
    '//// Read NES controller buttons
    nes_buttons := NES_Read_Gamepad

  'hang here until 'START' and '<SPACEBAR>' are released   
  repeat while (nes_buttons & CONSTANT( NES_START | (NES_START << 8) )) or (key.keystate(32)) and not((game_options & OPTION_MONKEY_IS_ALIVE) and (Rand_Range(1,100) == 1)) 
    '//// Read NES controller buttons
    nes_buttons := NES_Read_Gamepad


PUB Configure_Options | i, y, selected_option, string_ptr, num_options, menu_flags, key_repeat_ticks, current_option_mask, done

  selected_option := 0
  menu_flags      := 0
  done            := false

  '#IF EEPROM_GRAPHICS_ENABLE
  ' Get option screen
  'hdmf.EEPROM_Read(TILE_BUFFER, EEPROM_SCREEN_OPTIONS, CONSTANT((SCREEN_WIDTH * SCREEN_HEIGHT)>>3))
  'repeat until hdmf.EEPROM_IsDone
  '#ENDIF

  'Start music
  '#IF MUSIC_ENABLE  
  'hdmf.play_song(EEPROM_SONG_GOTHIC | hdmf#EEPROM_ASSET, hdmf#HDMF__LOOP)
  '#ENDIF

  gr.clear
  if DOUBLE_BUFFER
    gr.copy(@onscreen)

  repeat while not done

    'ClearBox(SCREEN_WIDTH-70,32,SCREEN_WIDTH-32,97)
    'gr.clear

    if not SFX_COG
      ' run the sounds effects periodically - not as good, but better than nothing!
      SFX_Manager

    '//// GET USER INPUT
    nes_buttons := NES_Read_Gamepad
    nes_buttons := ((nes_buttons & $ff00) >> 8) | nes_buttons  'Accept input from either gamepad  
      
    '//// Draw screen borders
    {
    gr.plot(-constant(SCREEN_WIDTH/2 -1),-constant(SCREEN_HEIGHT/2 -1))
    gr.line(constant(SCREEN_WIDTH/2 -1),-constant(SCREEN_HEIGHT/2 -1))
    gr.line(constant(SCREEN_WIDTH/2 -1),constant(SCREEN_HEIGHT/2 -1))
    gr.line(-constant(SCREEN_WIDTH/2 -1),constant(SCREEN_HEIGHT/2 -1))
    gr.line(-constant(SCREEN_WIDTH/2 -1),-constant(SCREEN_HEIGHT/2 -1)) 
    }

    '//// Draw titles
    gr.textmode(1,1,6,1)
    gr.colorwidth(1,0)
    '#IF !EEPROM_GRAPHICS_ENABLE
    { '<added by SCCuM preprocessor utility>
    gr.text(-20, 86, @title1_string)
    gr.text(-3, 73, @version_string)
    } '<added by SCCuM preprocessor utility>
    '#ENDIF
    
    'gr.text(-110, -51, @title3_string)
    'gr.text(-123, -64, @title4_string) 
    gr.text(-58, -111, @title_start_string) 
    
    gr.textmode(1,1,6,1)
    gr.text(-100, 83, @options_string)

    '//// Draw options states
    string_ptr := @option_strings
    y := 55
    num_options := 0
    current_option_mask := $00000001
    repeat while (byte[string_ptr] <> $ff)

      '//// Draw option name
      gr.text(-60, y, string_ptr)
      repeat while byte[string_ptr] <> $0
          ++string_ptr
      ++string_ptr

       '//// Draw option setting
      if (game_options & current_option_mask)
        gr.text(60, y, @on_string)
      else
        gr.text(60, y, @off_string)

      '//// Toggle option, if seleted
      '     Note: 'num_options' at this point is really the current option.
      if (nes_buttons & NES_A) and (selected_option == num_options) and not(menu_flags & MENU_FLAG__A_HELD)
        gr.colorwidth(0,0)
        gr.text(60, y, @off_string)
        gr.text(60, y, @on_string)
        gr.colorwidth(1,0)
        game_options ^=  current_option_mask  'Toggle the current option
        if (game_options & current_option_mask)
          gr.text(60, y, @on_string)
        else
          gr.text(60, y, @off_string)
       
      '//// Advance to next line
      y -= 13
      ++num_options
      current_option_mask <<= 1
    
    '//// Point to current option
    y :=55 -(13 * selected_option)
    gr.vec(90, y, $0020, ANG_180,player0_ship[0])

    '//// PROCESS MENU SELECTION, ON KEY DOWN EVENTS WITH KEY REPEATS
    if menu_flags & MENU_FLAG__UP_DOWN_HELD 
      ++key_repeat_ticks
      if not(nes_buttons & CONSTANT(NES_DOWN | NES_UP))
        menu_flags &= CONSTANT(!(MENU_FLAG__UP_DOWN_HELD | MENU_FLAG__REPEAT_GO_TIME_MET | MENU_FLAG__DO_REPEAT_CLICK))
      if not(menu_flags & MENU_FLAG__REPEAT_GO_TIME_MET) and (key_repeat_ticks => 8)
        menu_flags |= CONSTANT(MENU_FLAG__REPEAT_GO_TIME_MET | MENU_FLAG__DO_REPEAT_CLICK)
      if (menu_flags & MENU_FLAG__REPEAT_GO_TIME_MET) and (key_repeat_ticks => 2)
        menu_flags |= MENU_FLAG__DO_REPEAT_CLICK  
        
    if (not(menu_flags & MENU_FLAG__UP_DOWN_HELD)) or (menu_flags & MENU_FLAG__DO_REPEAT_CLICK)
      if nes_buttons & NES_UP  
        gr.colorwidth(0,0)
        y :=55 -(13 * selected_option)
        gr.vec(90, y, $0020, ANG_180, player0_ship[0])
        menu_flags |= MENU_FLAG__UP_DOWN_HELD
        if selected_option == 0
          selected_option := num_options - 1
        else
          --selected_option
      if nes_buttons & NES_DOWN  
        gr.colorwidth(0,0)
        y :=55 -(13 * selected_option)
        gr.vec(90, y, $0020, ANG_180, player0_ship[0])
        menu_flags |= MENU_FLAG__UP_DOWN_HELD  
        if selected_option == num_options - 1
          selected_option := 0
        else
          ++selected_option
      key_repeat_ticks := 0
      menu_flags &= !MENU_FLAG__DO_REPEAT_CLICK

    '//// COPY BITMAP TO DISPLAY
    if DOUBLE_BUFFER
      gr.copy(@onscreen)

    '//// PROCESS "START"
    if (nes_buttons & NES_START) or ((game_options & OPTION_MONKEY_IS_ALIVE) and (Rand_Range(1,100) == 1)) 
      done := true    

    '//// PROCESS BUTTON HELD FLAG
    if (nes_buttons & NES_A) 
       menu_flags |= MENU_FLAG__A_HELD
    else 
       menu_flags &= !MENU_FLAG__A_HELD

  'Stop playing music   
  '#IF MUSIC_ENABLE
  'hdmf.stop_song
  '#ENDIF

{
PUB ClearBox (x1,y1,x2,y2) | x,y,ptr,mask,i
'' This is a temprary fill routine to fill a box area of the screen with black
'' Only here because my hacked 2 color graphics driver does not currently support
'' a black "draw" color.

  {
  repeat x from x1 to x2
     repeat y from y1 to y2
        ptr := TILE_BUFFER + y * 4 + (((x & $ffe0) >> 5) * 4 * SCREEN_HEIGHT)
        mask :=  1 << (x & $1f)
        long[ptr] |= mask
  }
  
  'The proper version ran too slow.  This is just a hack to fill the required block quickly.
  ptr := TILE_BUFFER + y1 * 4 + (((x1 & $ffe0) >> 5) * 4 * SCREEN_HEIGHT)
  repeat i from ptr to ptr + (y2-y1) << 2
    long[i] := 0
    long[i + constant(4 * SCREEN_HEIGHT)] := 0 
}     
  

PUB Int_To_String(str, i) | t
   
  ' does an sprintf(str, "%05d", i); job
  str+=4
  repeat t from 0 to 4
    BYTE[str] := 48+(i // 10)
    i/=10
    str--

PUB Signed_Int_To_String(str, i) | t
   
  'Print the sign
  if (i<0)
    i := -i
    BYTE[str] := 45  ' ASCII '-'
  else
    BYTE[str] := 43  ' ASCII '+' 
    
  str+=5
  repeat t from 0 to 4
    BYTE[str] := 48+(i // 10)
    i/=10
    str--
    
PUB Fixed_To_String(str, fixed) | t , i
  ' does an sprintf(str, "%+05.04f", i); job

  'Print the sign
  if (fixed<0)
    fixed := -fixed
    BYTE[str] := 45  ' ASCII '-' 
  else
    BYTE[str] := 43  ' ASCII '+'     
    
  'Print the fractional portion
  str+=10
  i := ((fixed & $0000_FFFF) * 10000) ~>16 'Get the fractional part, and convert to 5 decimal digits of accuracy.
  repeat t from 0 to 3
    BYTE[str] := 48+(i // 10)
    i/=10
    str--
   
  'Print the decimal point
  BYTE[str] := 46  '46 = ASCII Decimal point
  str-- 
   
  'Print the decimal portion
  i:=fixed~>16
  repeat t from 0 to 4
    BYTE[str] := 48+(i // 10)
    i/=10
    str--

Pub Rand_Range(rstart, rend) : r_delta
  ' returns a random number from [rstart to rend] inclusive
  r_delta := rend - rstart + 1
  result := rstart + ((?random_var & $7FFFFFFF) // r_delta)
   
  return result

' ////////////////////////////////////////////////////////////////////

PUB SinCos(op, angle): xy               
  if (op==COS)
    angle+=$800   

  if angle & $1000 
    if angle & $800
      -angle
    angle |= $7000
    xy := -word[angle << 1]
  else
    if angle & $800
      -angle
    angle |= $7000
    xy := word[angle << 1]

' //////////////////////////////////////////////////////////////////////


PUB NES_Read_Gamepad : nes_bits        |  i

  ' //////////////////////////////////////////////////////////////////
  ' NES Game Paddle Read
  ' //////////////////////////////////////////////////////////////////       
  ' reads both gamepads in parallel encodes 8-bits for each in format
  ' right game pad #1 [15..8] : left game pad #0 [7..0]
  '
  ' set I/O ports to proper direction
  ' P3 = JOY_CLK      (4)
  ' P4 = JOY_SH/LDn   (5) 
  ' P5 = JOY_DATAOUT0 (6)
  ' P6 = JOY_DATAOUT1 (7)
  ' NES Bit Encoding
  '
  ' RIGHT  = %00000001
  ' LEFT   = %00000010
  ' DOWN   = %00000100
  ' UP     = %00001000
  ' START  = %00010000
  ' SELECT = %00100000
  ' B      = %01000000
  ' A      = %10000000
   
  ' step 1: set I/Os
  DIRA [3] := 1 ' output
  DIRA [4] := 1 ' output
  DIRA [5] := 0 ' input
  DIRA [6] := 0 ' input
   
  ' step 2: set clock and latch to 0
  OUTA [3] := 0 ' JOY_CLK = 0
  OUTA [4] := 0 ' JOY_SH/LDn = 0
  'Delay(1)
   
  ' step 3: set latch to 1
  OUTA [4] := 1 ' JOY_SH/LDn = 1
  'Delay(1)
   
  ' step 4: set latch to 0
  OUTA [4] := 0 ' JOY_SH/LDn = 0
   
  ' step 5: read first bit of each game pad
   
  ' data is now ready to shift out
  ' first bit is ready 
  nes_bits := 0
   
  ' left controller
  nes_bits := INA[5] | (INA[6] << 8)
   
  ' step 7: read next 7 bits
  repeat i from 0 to 6
   OUTA [3] := 1 ' JOY_CLK = 1
   'Delay(1)             
   OUTA [3] := 0 ' JOY_CLK = 0
   nes_bits := (nes_bits << 1)
   nes_bits := nes_bits | INA[5] | (INA[6] << 8)
   
   'Delay(1)             
  ' invert bits to make positive logic
  nes_bits := (!nes_bits & $FFFF)

  ' Zero bits if no controller present
  if (nes_bits & $ff00) == $ff00
    nes_bits &= $00ff
  if (nes_bits & $00ff) == $00ff
    nes_bits &= $ff00
   
  ' //////////////////////////////////////////////////////////////////
  ' End NES Game Paddle Read
  ' //////////////////////////////////////////////////////////////////

'///////////////////////////////////////////////////////////////////////
' DATA SECTION /////////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

VAR
   long sfx_flags

   word sfx_cnt_ship0_shot
   word sfx_cnt_ship1_shot
   byte sfx_cnt_shield
   byte shield_snd_ascending 

CON

   SFX_SND_SHIP0_SHOT_START = %00000000_00000001
   SFX_SND_SHIP0_SHOT       = %00000000_00000010  
   SFX_SND_SHIP1_SHOT_START = %00000000_00000100
   SFX_SND_SHIP1_SHOT       = %00000000_00001000
   SFX_SND_SHIP0_THRUST     = %00000000_00010000
   SFX_SND_SHIP1_THRUST     = %00000000_00100000  
   SFX_SND_SHIP0_SHIELD     = %00000000_01000000
   SFX_SND_SHIP1_SHIELD     = %00000000_10000000
   SFX_SND_EXPLODE          = %00000001_00000000
   SFX_SND_SHIELD_SUSTATIN  = %00000010_00000000
   SFX_SND_DAMAGE           = %00000100_00000000

   'if SFX_Manager has own cog:
   SFX_SHOT_SOUND_COUNT     = 511
   SFX_SHIELD_CNT_CYCLE     = 255
   SFX_SHOT_SCALE           = 2
   SFX_SHOT_DIVISOR         = 60

   'if SFX_Manager does not have own cog    
   SFX_SHOT_SOUND_COUNT_2   = 20
   SFX_SHIELD_CNT_CYCLE_2   = 10
   SFX_SHOT_SCALE_2         = 20
   SFX_SHOT_DIVISOR_2       = 15
   
PUB SFX_manager

'{
  repeat

    '-------------------
    ' SHOT SOUNDS
    '-------------------
    if (sfx_flags & SFX_SND_SHIP0_SHOT_START) <> 0
       'Start ship 0 shot sound  
       sfx_cnt_ship0_shot := 0
       sfx_flags &= !SFX_SND_SHIP0_SHOT_START
       sfx_flags |= SFX_SND_SHIP0_SHOT

    if (sfx_flags & SFX_SND_SHIP0_SHOT) <> 0
      'Play ship 0 shot sound
      if SFX_COG    
        hdmf.PlaySoundFM(5, hdmf#SHAPE_SAWTOOTH, $0400 - (sfx_cnt_ship0_shot * SFX_SHOT_SCALE) , CONSTANT( hdmf#SAMPLE_RATE / SFX_SHOT_DIVISOR ), 128, $2468_ACEF)
        if (++sfx_cnt_ship0_shot == SFX_SHOT_SOUND_COUNT)
          sfx_flags &= !SFX_SND_SHIP0_SHOT
      else
        hdmf.PlaySoundFM(5, hdmf#SHAPE_SAWTOOTH, $0400 - (sfx_cnt_ship0_shot * SFX_SHOT_SCALE_2) , CONSTANT( hdmf#SAMPLE_RATE / SFX_SHOT_DIVISOR_2 ), 128, $2468_ACEF)
        if (++sfx_cnt_ship0_shot == SFX_SHOT_SOUND_COUNT_2)
          sfx_flags &= !SFX_SND_SHIP0_SHOT

    if (sfx_flags & SFX_SND_SHIP1_SHOT_START) <> 0
      'Start ship 1 shot sound    
      sfx_cnt_ship1_shot := 0
      sfx_flags &= !SFX_SND_SHIP1_SHOT_START
      sfx_flags |= SFX_SND_SHIP1_SHOT

    if (sfx_flags & SFX_SND_SHIP1_SHOT) <> 0
      'Stop ship 1 shot sound
      if SFX_COG   
        hdmf.PlaySoundFM(4, hdmf#SHAPE_SAWTOOTH, $0520 - (sfx_cnt_ship1_shot * SFX_SHOT_SCALE) , CONSTANT( hdmf#SAMPLE_RATE / SFX_SHOT_DIVISOR ), 128, $2468_ACEF)
        if (++sfx_cnt_ship1_shot == SFX_SHOT_SOUND_COUNT)
          sfx_flags &= !SFX_SND_SHIP1_SHOT
      else
        hdmf.PlaySoundFM(4, hdmf#SHAPE_SAWTOOTH, $0520 - (sfx_cnt_ship1_shot * SFX_SHOT_SCALE_2) , CONSTANT( hdmf#SAMPLE_RATE / SFX_SHOT_DIVISOR_2 ), 128, $2468_ACEF)
        if (++sfx_cnt_ship1_shot == SFX_SHOT_SOUND_COUNT_2)
          sfx_flags &= !SFX_SND_SHIP1_SHOT

    '-------------------
    ' THRUST SOUNDS
    '-------------------
    if (sfx_flags & (SFX_SND_SHIP0_THRUST | SFX_SND_SHIP1_THRUST)) <> 0
      'Play thrust sound
      hdmf.PlaySoundFM(3, hdmf#SHAPE_NOISE, $080 + Rand_Range(0,$030) , CONSTANT( hdmf#SAMPLE_RATE / 2 ), 110, $2468_ACEF)
      

    '-------------------
    ' SHIELD SOUNDS
    '-------------------
    if (sfx_flags & (SFX_SND_SHIP0_SHIELD | SFX_SND_SHIP1_SHIELD | SFX_SND_SHIELD_SUSTATIN)) <> 0
      'play shield sound
      hdmf.PlaySoundFM(2, hdmf#SHAPE_SINE, $255 + sfx_cnt_shield * 2, CONSTANT( hdmf#SAMPLE_RATE / 8 ), 80, $2468_ACEF)

      'set sustain, so that shield sound always ends and bottom of pitch warble cycle
      sfx_flags |= SFX_SND_SHIELD_SUSTATIN
      
      'Model shield sound warble as a triangle shaped pitch change
      if shield_snd_ascending
        ++sfx_cnt_shield
        if (SFX_COG and (sfx_cnt_shield => SFX_SHIELD_CNT_CYCLE)) or (not SFX_COG and (sfx_cnt_shield => SFX_SHIELD_CNT_CYCLE_2))
          shield_snd_ascending := false
      else
        -- sfx_cnt_shield
        if(sfx_cnt_shield == 0)
          shield_snd_ascending := true
          'End sustain.  Shield sound will stop if shields are currently inactive. (This forces
          'the shield sound to always end at the bottom of the pitch warble cycle).
          sfx_flags &= !SFX_SND_SHIELD_SUSTATIN 
    else
      sfx_cnt_shield := 0
      shield_snd_ascending := true       

    '-------------------
    ' EXPLOSION SOUNDS
    '-------------------      
    if (sfx_flags & SFX_SND_EXPLODE) <> 0
      'Play explosion sound
      hdmf.PlaySoundFM(3, hdmf#SHAPE_NOISE, $100 , CONSTANT( hdmf#SAMPLE_RATE / 1 ), 110, $2468_ACEF)
      sfx_flags &= !SFX_SND_EXPLODE

    '-------------------
    ' DAMAGE SOUNDS
    '-------------------      
    if (sfx_flags & SFX_SND_DAMAGE) <> 0
      'Play explosion sound
      hdmf.PlaySoundFM(3, hdmf#SHAPE_NOISE, $200 , CONSTANT( hdmf#SAMPLE_RATE / 2 ), 110, $2468_ACEF)
      sfx_flags &= !SFX_SND_DAMAGE

  until not SFX_COG
      
'}

PUB SFX_stop_ship_action_sounds (player)

  if player == 0
    sfx_flags &= !(SFX_SND_SHIP0_THRUST | SFX_SND_SHIP0_SHIELD)
  else
    sfx_flags &= !(SFX_SND_SHIP1_THRUST | SFX_SND_SHIP1_SHIELD)  


DAT


' BLACK HOLES ///////////////////////////////////////////////////////////
{
' black hole initialization
black_x                  long 20,  80, -90 ' x position of each black hole
black_y                  long 50, -60, -10 ' y position of each black hole
black_gravity            long  8,  4,  2   ' single value that is "lumped" parameter that relates to the final "pull"
                                           ' of the black hole relative to the ship's constant mass
}

' STRING STORAGE //////////////////////////////////////////////////////
{
score_string            byte    "Score",0               'text
hiscore_string          byte    "High",0                'text
ships_string            byte    "Ships",0               'text
}

' TV PARAMETERS FOR DRIVER /////////////////////////////////////////////

vgaparams               long    0               'status
                        long    1               'enable
                        long    %010_111        'pins
                        long    mode
                        long    0               'screen
                        long    0               'colors
                        long    x_tiles         'hc
                        long    y_tiles         'vc
                        long    1               'hx
                        long    1               'vx
                        long    0               'ho
                        long    0               'vo
                        long    hd
                        long    hf
                        long    hs
                        long    hb
                        long    vd
                        long    vf
                        long    vs
                        long    vb
                        long    px


' POLYGON OBJECTS ///////////////////////////////////////////////////////
  
player0_ship0
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    80

                        word    VECTOR_DRAW + ANG_120+ANG_10 + ANG_10           '2
                        word    70

                        word    VECTOR_DRAW + ANG_120+ANG_10                    '3
                        word    20

                        word    VECTOR_DRAW + ANG_240-ANG_10                    '4
                        word    20

                        word    VECTOR_DRAW + ANG_180+ANG_10                    '5
                        word    34

                        word    VECTOR_DRAW + ANG_180-ANG_10                    '6
                        word    34
                        
                        word    VECTOR_DRAW + ANG_120+ANG_10                    '7
                        word    20

                        word    VECTOR_MOVE + ANG_240-ANG_10                    '8
                        word    20

                        word    VECTOR_DRAW + ANG_240-ANG_10 - ANG_10           '9
                        word    70

                        word    VECTOR_DRAW + ANG_0                             '10
                        word    80

                        word    VECTOR_MOVE + ANG_90                            '11
                        word    25

                        word    VECTOR_DRAW + ANG_0                             '12
                        word    28

                        word    VECTOR_DRAW + ANG_270                           '13
                        word    25

                        word    VECTOR_END
                        word    0

player0_ship1
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    80

                        'word    VECTOR_DRAW + ANG_120+ANG_10 + ANG_10           '2
                        'word    70

                        word    VECTOR_DRAW + ANG_90                            '2i
                        word    25

                        word    VECTOR_DRAW + ANG_120+ANG_10                    '3
                        word    20

                        word    VECTOR_DRAW + ANG_240-ANG_10                    '4
                        word    20

                        word    VECTOR_DRAW + ANG_180+ANG_10                    '5
                        word    34

                        word    VECTOR_DRAW + ANG_180-ANG_10                    '6
                        word    34
                        
                        word    VECTOR_DRAW + ANG_120+ANG_10                    '7
                        word    20

                        word    VECTOR_MOVE + ANG_240-ANG_10                    '8
                        word    20

                        word    VECTOR_DRAW + ANG_240-ANG_10 - ANG_10           '9
                        word    70

                        word    VECTOR_DRAW + ANG_0                             '10
                        word    80

                        word    VECTOR_MOVE + ANG_90                            '11
                        word    25

                        word    VECTOR_DRAW + ANG_0                             '12
                        word    28

                        word    VECTOR_DRAW + ANG_270                           '13
                        word    25

                        word    VECTOR_END
                        word    0

player0_ship2 
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    80

                        word    VECTOR_DRAW + ANG_120+ANG_10 + ANG_10           '2
                        word    70

                        word    VECTOR_DRAW + ANG_120+ANG_10                    '3
                        word    20

                        word    VECTOR_DRAW + ANG_240-ANG_10                    '4
                        word    20

                        word    VECTOR_DRAW + ANG_180+ANG_10                    '5
                        word    34

                        word    VECTOR_DRAW + ANG_180-ANG_10                    '6
                        word    34
                        
                        word    VECTOR_DRAW + ANG_120+ANG_10                    '7
                        word    20

                        word    VECTOR_MOVE + ANG_240-ANG_10                    '8
                        word    20

                        'word    VECTOR_DRAW + ANG_240-ANG_10 - ANG_10           '9
                        'word    70

                        word    VECTOR_DRAW + ANG_270                           '9i
                        word    25
                        
                        word    VECTOR_DRAW + ANG_0                             '10
                        word    80

                        word    VECTOR_MOVE + ANG_90                            '11
                        word    25

                        word    VECTOR_DRAW + ANG_0                             '12
                        word    28

                        word    VECTOR_DRAW + ANG_270                           '13
                        word    25

                        word    VECTOR_END
                        word    0

player0_ship3
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    80

                        'word    VECTOR_DRAW + ANG_120+ANG_10 + ANG_10           '2
                        'word    70

                        word    VECTOR_DRAW + ANG_90                            '2i
                        word    25

                        word    VECTOR_DRAW + ANG_120+ANG_10                    '3
                        word    20

                        word    VECTOR_DRAW + ANG_240-ANG_10                    '4
                        word    20

                        'word    VECTOR_DRAW + ANG_180+ANG_10                    '5
                        'word    34

                        'word    VECTOR_DRAW + ANG_180-ANG_10                    '6
                        'word    34
                        
                        'word    VECTOR_DRAW + ANG_120+ANG_10                    '7
                        'word    20

                        word    VECTOR_MOVE + ANG_240-ANG_10                    '8
                        word    20

                        'word    VECTOR_DRAW + ANG_240-ANG_10 - ANG_10           '9
                        'word    70

                        word    VECTOR_DRAW + ANG_270                           '9i
                        word    25
                        
                        word    VECTOR_DRAW + ANG_0                             '10
                        word    80

                        word    VECTOR_MOVE + ANG_90                            '11
                        word    25

                        word    VECTOR_DRAW + ANG_0                             '12
                        word    28

                        word    VECTOR_DRAW + ANG_270                           '13
                        word    25

                        word    VECTOR_END
                        word    0

player1_ship0
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    36

                        word    VECTOR_DRAW + ANG_60                            '2
                        word    32

                        word    VECTOR_DRAW + ANG_120                           '3
                        word    32

                        word    VECTOR_DRAW + ANG_180                           '4
                        word    36
                                                                                
                        word    VECTOR_DRAW + ANG_240                           '5
                        word    32

                        word    VECTOR_DRAW + ANG_300                           '6
                        word    32

                        word    VECTOR_DRAW + ANG_0                             '7
                        word    32

                        word    VECTOR_MOVE + ANG_0                             '8
                        word    8

                        word    VECTOR_DRAW + ANG_0                             '9
                        word    38

                        word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10         '10
                        word    30

                        word    VECTOR_DRAW + ANG_180                           '11
                        word    80

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10         '12
                        word    30

                        word    VECTOR_MOVE + ANG_180 - ANG_10                  '13
                        word    48

                        word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_10 - ANG_5 '14
                        word    75

                        word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10 - ANG_10 - ANG_10 - ANG_1       '15 
                        word    58

                        word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_5 - ANG_1 - ANG_1 - ANG_1          '16
                        word    120

                        word    VECTOR_MOVE + ANG_180 + ANG_10                  '17
                        word    48

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10 + ANG_5 '18
                        word    75

                        word    VECTOR_MOVE + ANG_180 + ANG_10 + ANG_10 + ANG_10 + ANG_10 + ANG_1       '19
                        word    58

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_5 + ANG_1 + ANG_1 + ANG_1          '20
                        word    120

                        word    VECTOR_END
                        word    0                                   


player1_ship1
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    36

                        word    VECTOR_DRAW + ANG_60                            '2
                        word    32

                        word    VECTOR_DRAW + ANG_120                           '3
                        word    32

                        word    VECTOR_DRAW + ANG_180                           '4
                        word    36
                                                                                
                        word    VECTOR_DRAW + ANG_240                           '5
                        word    32

                        word    VECTOR_DRAW + ANG_300                           '6
                        word    32

                        word    VECTOR_DRAW + ANG_0                             '7
                        word    32

                        word    VECTOR_MOVE + ANG_0                             '8
                        word    8

                        word    VECTOR_DRAW + ANG_0                             '9
                        word    38

                        word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10         '10
                        word    30

                        word    VECTOR_DRAW + ANG_180                           '11
                        word    80

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10         '12
                        word    30

                        'word    VECTOR_MOVE + ANG_180 - ANG_10                  '13
                        'word    48

                        'word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_10 - ANG_5 '14
                        'word    75

                        'word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10 - ANG_10 - ANG_10 - ANG_1       '15 
                        'word    58

                        'word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_5 - ANG_1 - ANG_1 - ANG_1          '16
                        'word    120

                        word    VECTOR_MOVE + ANG_180 + ANG_10                  '17
                        word    48

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10 + ANG_5 '18
                        word    75

                        word    VECTOR_MOVE + ANG_180 + ANG_10 + ANG_10 + ANG_10 + ANG_10 + ANG_1       '19
                        word    58

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_5 + ANG_1 + ANG_1 + ANG_1          '20
                        word    120

                        word    VECTOR_END
                        word    0                                    

player1_ship2
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    36

                        word    VECTOR_DRAW + ANG_60                            '2
                        word    32

                        word    VECTOR_DRAW + ANG_120                           '3
                        word    32

                        word    VECTOR_DRAW + ANG_180                           '4
                        word    36
                                                                                
                        word    VECTOR_DRAW + ANG_240                           '5
                        word    32

                        word    VECTOR_DRAW + ANG_300                           '6
                        word    32

                        word    VECTOR_DRAW + ANG_0                             '7
                        word    32

                        word    VECTOR_MOVE + ANG_0                             '8
                        word    8

                        word    VECTOR_DRAW + ANG_0                             '9
                        word    38

                        word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10         '10
                        word    30

                        word    VECTOR_DRAW + ANG_180                           '11
                        word    80

                        word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10         '12
                        word    30

                        word    VECTOR_MOVE + ANG_180 - ANG_10                  '13
                        word    48

                        word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_10 - ANG_5 '14
                        word    75

                        word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10 - ANG_10 - ANG_10 - ANG_1       '15 
                        word    58

                        word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_5 - ANG_1 - ANG_1 - ANG_1          '16
                        word    120

                        'word    VECTOR_MOVE + ANG_180 + ANG_10                  '17
                        'word    48

                        'word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10 + ANG_5 '18
                        'word    75

                        'word    VECTOR_MOVE + ANG_180 + ANG_10 + ANG_10 + ANG_10 + ANG_10 + ANG_1       '19
                        'word    58

                        'word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_5 + ANG_1 + ANG_1 + ANG_1          '20
                        'word    120

                        word    VECTOR_END
                        word    0                                    


player1_ship3
                        word    VECTOR_MOVE + ANG_0                             '1             
                        word    36

                        word    VECTOR_DRAW + ANG_60                            '2
                        word    32

                        word    VECTOR_DRAW + ANG_120                           '3
                        word    32

                        word    VECTOR_DRAW + ANG_180                           '4
                        word    36
                                                                                
                        word    VECTOR_DRAW + ANG_240                           '5
                        word    32

                        word    VECTOR_DRAW + ANG_300                           '6
                        word    32

                        word    VECTOR_DRAW + ANG_0                             '7
                        word    32

                        word    VECTOR_MOVE + ANG_0                             '8
                        word    8

                        word    VECTOR_DRAW + ANG_0                             '9
                        word    38

                        'word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10         '10
                        'word    30

                        'word    VECTOR_DRAW + ANG_180                           '11
                        'word    80

                        'word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10         '12
                        'word    30

                        'word    VECTOR_MOVE + ANG_180 - ANG_10                  '13
                        'word    48

                        'word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_10 - ANG_5 '14
                        'word    75

                        'word    VECTOR_MOVE + ANG_180 - ANG_10 - ANG_10 - ANG_10 - ANG_10 - ANG_1       '15 
                        'word    58

                        'word    VECTOR_DRAW + ANG_180 - ANG_10 - ANG_5 - ANG_1 - ANG_1 - ANG_1          '16
                        'word    120

                        'word    VECTOR_MOVE + ANG_180 + ANG_10                  '17
                        'word    48

                        'word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_10 + ANG_5 '18
                        'word    75

                        'word    VECTOR_MOVE + ANG_180 + ANG_10 + ANG_10 + ANG_10 + ANG_10 + ANG_1       '19
                        'word    58

                        'word    VECTOR_DRAW + ANG_180 + ANG_10 + ANG_5 + ANG_1 + ANG_1 + ANG_1          '20
                        'word    120

                        word    VECTOR_END
                        word    0                                    


vectors_shield          word    VECTOR_MOVE + ANG_0                         ' Octagonal shield          
                        word    70

                        word    VECTOR_DRAW + ANG_45
                        word    70
                        
                        word    VECTOR_MOVE + ANG_90
                        word    70

                        word    VECTOR_DRAW + ANG_90+ANG_45 
                        word    70

                        word    VECTOR_MOVE + ANG_180  
                        word    70

                        word    VECTOR_DRAW + ANG_180+ANG_45   
                        word    70

                        word    VECTOR_MOVE + ANG_270  
                        word    70

                        word    VECTOR_DRAW + ANG_270+ANG_45   
                        word    70

                        word    VECTOR_MOVE + ANG_0             
                        word    70

                        word    VECTOR_END
                        word    0


vectors_rock            word    VECTOR_MOVE + ANG_0 
                        word    70

                        word    VECTOR_DRAW + (ANG_1 * 50)
                        word    75

                        word    VECTOR_DRAW + (ANG_1 * 100)
                        word    70

                        word    VECTOR_DRAW + (ANG_1 * 140)
                        word    50

                        word    VECTOR_DRAW + (ANG_1 * 170)
                        word    75

                        word    VECTOR_DRAW + (ANG_1 * 200)
                        word    65

                        word    VECTOR_DRAW + (ANG_1 * 250)
                        word    50

                        word    VECTOR_DRAW + (ANG_1 * 320)
                        word    70

                        word    VECTOR_DRAW + (ANG_1 * 350)
                        word    40

                        word    VECTOR_DRAW + ANG_0
                        word    70

                        word    VECTOR_END
                        word    0


{
vectors_bfg             word    VECTOR_MOVE + ANG_0 
                        word    BFG_SPOKE_SIZE

                        word    VECTOR_DRAW + (ANG_1 * 60)
                        word    BFG_HUB_SIZE

                        word    VECTOR_DRAW + (ANG_1 * 120)
                        word    BFG_SPOKE_SIZE

                        word    VECTOR_DRAW + (ANG_1 * 180)
                        word    BFG_HUB_SIZE

                        word    VECTOR_DRAW + (ANG_1 * 240)
                        word    BFG_SPOKE_SIZE

                        word    VECTOR_DRAW + (ANG_1 * 300)
                        word    BFG_HUB_SIZE

                        word    VECTOR_DRAW + ANG_0
                        word    BFG_SPOKE_SIZE

                        word    VECTOR_END
                        word    0                    
                        
}
fixed_pt_string_0        byte    "-00000.0000",0    'text
int_string_0             byte    "-00000",0         'text
game_over_string         byte    "GAME OVER",0      'text
game_over_restart        byte    "PRESS <START>",0  'text
'debug_string             byte    "DEBUG",0          'text

title1_string            byte    "SPACEWAR!",0                  'text
version_string           byte    "3.2 (HIRES VGA)",0            'text
title3_string            byte    "WRITTEN BY ERIC MOYER",0      'text
title4_string            byte    "VGA MODS BY ROSS HIGSON",0    'text
title_start_string       byte    "PRESS <START> TO BEGIN",0     'text 

options_string           byte    "OPTIONS - UP/DOWN SELECTS, 'A' CHANGES",0        'text

option_strings           byte    "BLACK HOLE",0     'text
                         byte    "SHOT GRAVITY",0   'text
                         byte    "INERTIAL SHOTS",0 'text
                         byte    "SHIELDS",0        'text
                         byte    "FUEL",0           'text
                         byte    "TWINKLE",0        'text
                         byte    "HYPERSPACE",0     'text
                         byte    "SHIP DAMAGE",0    'text
                         byte    "RANDOM ROCKS",0   'text
                         byte    "BOUNCE BACK",0    'text
                         byte    "NEGATIVE GRAVITY",0 'text
                         byte    "EXPANDED UNIVERSE",0 'text
                         byte    $ff                'end of option string list  
'                         byte    "BFG",0            'text
                         
                         
on_string                byte    "ON",0              'text
off_string               byte    "OFF",0             'text

'#IF MUSIC_ENABLE
{ 
_song_mario
        byte    $02, $EF, $AD, $79, $35          'Instrument 0
        byte    $FF                              'End of Instrument list
        byte    $00, $00, $49, $0D, $0F
        byte    $00, $20, $93, $0D, $0F
        byte    $00, $42, $93, $0D, $0F
        byte    $10, $61, $72, $0D, $0F
        byte    $00, $80, $93, $0D, $0F
        byte    $00, $A0, $49, $0D, $0F
        byte    $00, $02, $93, $0D, $0F
        byte    $1F, $20, $49, $0D, $0F
        byte    $00, $41, $72, $0D, $0F
        byte    $00, $02, $93, $0D, $0F
        byte    $00, $60, $93, $0D, $0F
        byte    $1F, $81, $72, $0D, $0F
        byte    $00, $A0, $93, $0D, $0F
        byte    $00, $00, $49, $0D, $0F
        byte    $00, $22, $0B, $0D, $0F
        byte    $10, $41, $72, $0D, $0F
        byte    $00, $60, $49, $0D, $0F
        byte    $00, $00, $93, $0D, $0F
        byte    $00, $22, $93, $0D, $0F
        byte    $1F, $83, $10, $0D, $10
        byte    $00, $A1, $EE, $0D, $10
        byte    $00, $01, $88, $0D, $10
        byte    $00, $20, $C4, $0D, $10
        byte    $00, $40, $62, $0D, $10
        byte    $3F, $61, $88, $0D, $10
        byte    $00, $00, $C4, $0D, $10
        byte    $00, $20, $62, $0D, $10
        byte    $3E, $41, $4A, $0D, $0F
        byte    $00, $80, $C4, $0D, $0F
        byte    $00, $A0, $83, $0D, $0F
        byte    $00, $02, $0B, $0D, $0F
        byte    $2F, $21, $06, $0D, $0F
        byte    $00, $60, $A5, $0D, $0F
        byte    $00, $00, $83, $0D, $0F
        byte    $00, $41, $88, $0D, $0F
        byte    $2F, $80, $83, $0D, $0F
        byte    $00, $A1, $4A, $0D, $0F
        byte    $00, $00, $C4, $0D, $0F
        byte    $2F, $21, $06, $0D, $0F
        byte    $00, $41, $B8, $0D, $0F
        byte    $00, $60, $AF, $0D, $0F
        byte    $1F, $01, $EE, $0D, $0F
        byte    $00, $81, $26, $0D, $0F
        byte    $00, $A0, $C4, $0D, $0F
        byte    $1F, $21, $D2, $0D, $0F
        byte    $00, $41, $15, $0D, $0F
        byte    $00, $60, $B9, $0D, $0F
        byte    $10, $01, $B8, $0D, $0F
        byte    $00, $81, $06, $0D, $0F
        byte    $00, $A0, $AF, $0D, $0F
        byte    $1F, $21, $06, $0D, $14
        byte    $00, $41, $88, $0D, $14
        byte    $00, $60, $A5, $0D, $14
        byte    $15, $01, $88, $0D, $14
        byte    $00, $81, $06, $0D, $14
        byte    $00, $A0, $C4, $0D, $14
        byte    $00, $22, $93, $0D, $14
        byte    $15, $41, $EE, $0D, $14
        byte    $00, $61, $06, $0D, $14
        byte    $00, $03, $10, $0D, $14
        byte    $00, $21, $4A, $0D, $14
        byte    $14, $81, $5D, $0D, $0F
        byte    $00, $A0, $DC, $0D, $0F
        byte    $00, $03, $70, $0D, $0F
        byte    $00, $22, $0B, $0D, $0F
        byte    $20, $41, $B8, $0D, $0F
        byte    $00, $61, $26, $0D, $0F
        byte    $00, $02, $BA, $0D, $0F
        byte    $0F, $21, $EE, $0D, $1F
        byte    $00, $83, $10, $0D, $1F
        byte    $00, $A1, $4A, $0D, $1F
        byte    $20, $01, $B8, $0D, $0F
        byte    $00, $42, $93, $0D, $0F
        byte    $00, $60, $DC, $0D, $0F
        byte    $00, $21, $06, $0D, $0F
        byte    $1F, $82, $0B, $0D, $0F
        byte    $00, $A1, $4A, $0D, $0F
        byte    $00, $00, $DC, $0D, $0F
        byte    $10, $22, $4B, $0D, $0F
        byte    $00, $41, $5D, $0D, $0F
        byte    $00, $60, $F7, $0D, $0F
        byte    $0F, $01, $26, $0D, $0F
        byte    $00, $81, $EE, $0D, $0F
        byte    $00, $A0, $C4, $0D, $0F
        byte    $2F, $22, $0B, $0D, $0F
        byte    $00, $41, $4A, $0D, $0F
        byte    $00, $60, $C4, $0D, $0F
        byte    $00, $00, $83, $0D, $0F
        byte    $2F, $81, $06, $0D, $0F
        byte    $00, $A1, $88, $0D, $0F
        byte    $00, $00, $A5, $0D, $0F
        byte    $00, $20, $83, $0D, $0F
        byte    $2F, $41, $4A, $0D, $0F
        byte    $00, $60, $C4, $0D, $0F
        byte    $00, $00, $83, $0D, $0F
        byte    $2F, $21, $B8, $0D, $0F
        byte    $00, $81, $06, $0D, $0F
        byte    $00, $A0, $AF, $0D, $0F
        byte    $1F, $01, $26, $0D, $0F
        byte    $ff
'#ENDIF
}